/*********************************************************************************
* Copyright (C) 2015-2017 Alexey V. Akimov
*
* This file is distributed under the terms of the GNU General Public License
* as published by the Free Software Foundation, either version 2 of
* the License, or (at your option) any later version.
* See the file LICENSE in the root directory of this distribution
* or <http://www.gnu.org/licenses/>.
*
*********************************************************************************/

#include "A_coefficients.h"
#include "../math_specialfunctions/libspecialfunctions.h"


/// liblibra namespace
namespace liblibra{

using namespace libspecialfunctions;

namespace libmolint{


void Aux_F1(double rhoA, double rhoB, double* f,int mu_max){

  if(rhoA==rhoB){
    double exp_ = exp(-rhoA);
    for(int mu=0;mu<=mu_max;mu++){ f[mu] = exp_; }
  }
  else{
    double drho = (rhoA-rhoB);
    double drho2 = (1.0/(drho*drho));
    double exp_a = exp(-rhoA);
    double exp_b = exp(-rhoB);
    
    double f_minus1 = 0.5*(exp_b+exp_a);    
    f[0] = (exp_b - exp_a) / drho ;

    if(mu_max>=1){
      f[1] = 12.0*drho2*(f_minus1 - f[0]);

      for(int mu=2;mu<=mu_max;mu++){
        f[mu] = 4*(2*mu+1)*(2*mu-1)*drho2*(f[mu-2]-f[mu-1]);
      }

    }// mu_max>=1
    
  }// else: rhoA != rhoB

}// Aux_F1


double A_coefficient_general(int u,int v, int na, int nb, int la, int lb, int m){
/*
   Implements the function Eq.(28) from:

   Silver, D.M.; Ruedenberg, K.  J. Chem. Phys. 1968, 49, 4301


   Notation notes:  format: my_notation  = text_notation
   u = mu;  v = nu;                            
   na = n; nb = n_prime;
   la = l; lb = l_prime;   
   alp = alpha; alpp = alpha_prime;
   beta= beta;  betap = beta_prime;

*/

  double res = 0.0;


  int beta1 = max(max(0, u-nb-la), v-nb-la+m);
  int beta2 = min(na-la,v);
//  cout<<"beta1= "<<beta1<<" beta2= "<<beta2<<endl;

  int delta = 1 + u + min(v+la+lb, na+nb);

  for(int beta=beta1;beta<=beta2;beta++){

    int beta1p = max(max(0,u-beta-la-lb),v-beta-la-lb+m);
    int beta2p = min(nb-lb,v-beta);
//    cout<<"beta1p= "<<beta1p<<" beta2p= "<<beta2p<<endl;

    double sum0 = 0.0;

    for(int betap=beta1p;betap<=beta2p;betap++){

      double sum = 0.0;
      int alp1 = m + max(0,v-beta-betap-lb);

//      cout<<"alp1= "<<alp1<<" la= "<<la<<endl;

      for(int alp=alp1;alp<=la;alp++){

        double sum1 = 0.0;
        int alp1p = m + max(0,v-beta-betap-alp);

        for(int alpp=alp1p;alpp<=lb;alpp++){

          double sum2 = 0.0;

          for(int lamb=0;lamb<=u;lamb++){

//            cout<<"beta= "<<beta<<" betap= "<<betap<<" alp= "<<alp<<" alpp= "<<alpp<<" lamb= "<<lamb<<endl;
            
            sum2 += pow(-1,lamb)*BINOM(lamb,alp-alpp+beta+lb+lamb)*BINOM(u-lamb,alpp-alp+betap+la+u-lamb)*BINOM(lamb,u);
          }

          sum1 += pow(-1,alpp)*BINOM(alpp,lb)*BINOM(alpp-m,lb)*BINOM(v-beta-betap,alp+alpp-m)
                  *FACTORIAL(alp-alpp+beta+lb)*FACTORIAL(alpp-alp+betap+la)*sum2;
        }
        
        sum += pow(-1,alp)*BINOM(alp,la)*BINOM(alp-m,la)*sum1;

      }// for alp

      sum0 += BINOM(betap,nb-lb)*(FACTORIAL(delta)/FACTORIAL(u+beta+betap+la+lb+1)) * sum;

    }// for betap

    res += BINOM(beta,na-la)*sum0;

  }// for beta


  double prefact1 = sqrt((2*la+1)*(2*lb+1)*BINOM(m,la+m)*BINOM(m,lb+m)/(BINOM(m,la)*BINOM(m,lb)*BINOM(na,2*na)*BINOM(nb,2*nb)));
  prefact1 *= pow(-1,la+lb)*FACTORIAL(na+nb-v)/(FACTORIAL(na)*FACTORIAL(nb)*FACTORIAL(delta)*BINOM(u,2*u));

  res *= prefact1;

  return res;

}


void generate_coefficients(){
/* Use the following snippet to generate A coeffiecients in tabular form, to avoid explicit calculation 
   The function implementing the resulting table is below   
*/


  for(int na=0;na<=3;na++){
    if(na==0){  cout<<"if(na=="<<na<<"){\n"; }
    else{   cout<<"else if(na=="<<na<<"){\n"; }

    for(int nb=0;nb<=3;nb++){
      if(nb==0){  cout<<"  if(nb=="<<nb<<"){\n"; }
      else{   cout<<"  else if(nb=="<<nb<<"){\n"; }
      

      for(int la=0;la<=na;la++){
        if(la==0){  cout<<"    if(la=="<<la<<"){\n"; }
        else{   cout<<"    else if(la=="<<la<<"){\n"; }

        for(int lb=0;lb<=nb;lb++){
          if(lb==0){  cout<<"      if(lb=="<<lb<<"){\n"; }
          else{   cout<<"      else if(lb=="<<lb<<"){\n"; }


          for(int m=0;m<=min(la,lb);m++){
            if(m==0){  cout<<"        if(abs(m)=="<<m<<"){\n"; }
            else{   cout<<"        else if(abs(m)=="<<m<<"){\n"; }



            for(int u=0;u<=na+nb;u++){
              int v1 = max(max(0,abs(la-lb)-u),u-(la+lb));
              int v2 = na+nb-m;

              for(int v=v1;v<=v2;v++){

                double res = A_coefficient_general(u,v,na,nb,la,lb,m);

                cout<<std::setprecision(25) <<"          A["<<u<<"]["<<v<<"] = "<<res<<";\n"; 

              }//for v
            }// for u

            cout<<"        } // |m|="<<m<<"\n";
          }// for m


          cout<<"      } // lb="<<lb<<"\n";
        }// for lb

        cout<<"    } // la="<<la<<"\n";
      }// for la

      cout<<"  } // nb="<<nb<<"\n";
    }// for nb  

    cout<<"} // na="<<na<<"\n";
  }// for na

}


void A_coefficients_fast(int u,int v, int na, int nb, int la, int lb, int m,double** A){
/*
  This is a faster (tabular) form of the A_coefficient_general function described above

*/
if(na==0){
  if(nb==0){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 1.000000000000000000000000;
        } // |m|=0
      } // lb=0
    } // la=0
  } // nb=0
  else if(nb==1){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.7071067811865475727373109;
          A[0][1] = 0.3535533905932737863686555;
          A[1][1] = 0.05892556509887896670774055;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.6123724356957944703339081;
          A[1][0] = 0.2041241452319314808594442;
          A[1][1] = 0.1020620726159657404297221;
        } // |m|=0
      } // lb=1
    } // la=0
  } // nb=1
  else if(nb==2){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.4082482904638630172300395;
          A[0][1] = 0.2041241452319315086150198;
          A[0][2] = 0.06804138174397716953833992;
          A[1][1] = 0.03402069087198858476916996;
          A[1][2] = 0.01701034543599429238458498;
          A[2][2] = 0.0005670115145331430578021226;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.2357022603955158668309622;
          A[0][2] = 0.1178511301977579334154811;
          A[1][0] = 0.1178511301977579334154811;
          A[1][1] = 0.05892556509887896670774055;
          A[1][2] = 0.02946278254943948335387027;
          A[2][1] = 0.001964185503295965470188511;
          A[2][2] = 0.0009820927516479827350942555;
        } // |m|=0
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][2] = 0.1521451548625461502872724;
          A[1][1] = 0.07607257743127307514363622;
          A[1][2] = 0.03803628871563653757181811;
          A[2][0] = 0.01521451548625461433483785;
          A[2][1] = 0.007607257743127307167418927;
          A[2][2] = 0.001267876290521218005796777;
        } // |m|=0
      } // lb=2
    } // la=0
  } // nb=2
  else if(nb==3){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.2236067977499789638518024;
          A[0][1] = 0.1118033988749894819259012;
          A[0][2] = 0.03726779962499649628826504;
          A[0][3] = 0.009316949906249124072066259;
          A[1][1] = 0.01863389981249824814413252;
          A[1][2] = 0.009316949906249124072066259;
          A[1][3] = 0.002795084971874737134883704;
          A[2][2] = 0.0003105649968749707987882014;
          A[2][3] = 0.0001552824984374853993941007;
          A[3][3] = 2.218321406249791226308193e-06;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.1075828707279838092336632;
          A[0][2] = 0.05379143536399190461683162;
          A[0][3] = 0.01613743060919757207893888;
          A[1][0] = 0.06454972243679028831575550;
          A[1][1] = 0.03227486121839514415787775;
          A[1][2] = 0.01506160190191773232126771;
          A[1][3] = 0.004841229182759271103264620;
          A[2][1] = 0.001434438276373117436210403;
          A[2][2] = 0.0007172191381865587181052013;
          A[2][3] = 0.0002689571768199595057369233;
          A[3][2] = 7.684490766284557790788065e-06;
          A[3][3] = 3.842245383142278895394033e-06;
        } // |m|=0
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][2] = 0.04166666666666666435370203;
          A[0][3] = 0.02083333333333333217685102;
          A[1][1] = 0.03333333333333333287074041;
          A[1][2] = 0.01666666666666666643537020;
          A[1][3] = 0.006250000000000000346944695;
          A[2][0] = 0.008333333333333333217685102;
          A[2][1] = 0.004166666666666666608842551;
          A[2][2] = 0.001388888888888888941894328;
          A[2][3] = 0.0003472222222222222354735821;
          A[3][1] = 5.952380952380952375789513e-05;
          A[3][2] = 2.976190476190476187894757e-05;
          A[3][3] = 4.960317460317460313157928e-06;
        } // |m|=0
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][3] = 0.02465033242958173179126824;
          A[1][2] = 0.01479019945774904046253972;
          A[1][3] = 0.007395099728874520231269862;
          A[2][1] = 0.004930066485916346531725996;
          A[2][2] = 0.002465033242958173265862998;
          A[2][3] = 0.0004108388738263621929071301;
          A[3][0] = 0.0007042952122737638522009807;
          A[3][1] = 0.0003521476061368819261004903;
          A[3][2] = 7.042952122737637708858177e-05;
          A[3][3] = 5.869126768948031424048481e-06;
        } // |m|=0
      } // lb=3
    } // la=0
  } // nb=3
} // na=0
else if(na==1){
  if(nb==0){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.7071067811865475727373109;
          A[0][1] = 0.3535533905932737863686555;
          A[1][1] = -0.05892556509887896670774055;
        } // |m|=0
      } // lb=0
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.6123724356957944703339081;
          A[1][0] = -0.2041241452319314808594442;
          A[1][1] = -0.1020620726159657404297221;
        } // |m|=0
      } // lb=0
    } // la=1
  } // nb=0
  else if(nb==1){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.08333333333333332870740406;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[2][2] = -0.001388888888888888941894328;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.2886751345948128655294340;
          A[0][2] = 0.1443375672974064327647170;
          A[1][0] = 0.2886751345948128655294340;
          A[1][1] = 0.1443375672974064327647170;
          A[1][2] = -0.000000000000000000000000;
          A[2][1] = -0.004811252243246881207805465;
          A[2][2] = -0.002405626121623440603902733;
        } // |m|=0
      } // lb=1
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.2886751345948128655294340;
          A[0][2] = 0.1443375672974064327647170;
          A[1][0] = -0.2886751345948128655294340;
          A[1][1] = -0.1443375672974064327647170;
          A[1][2] = -0.000000000000000000000000;
          A[2][1] = -0.004811252243246881207805465;
          A[2][2] = -0.002405626121623440603902733;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][0] = -1.000000000000000000000000;
          A[0][1] = -0.5000000000000000000000000;
          A[0][2] = 0.2500000000000000000000000;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[2][0] = -0.03333333333333333287074041;
          A[2][1] = -0.01666666666666666643537020;
          A[2][2] = -0.004166666666666666608842551;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[2][0] = -0.01666666666666666643537020;
          A[2][1] = -0.008333333333333333217685102;
        } // |m|=1
      } // lb=1
    } // la=1
  } // nb=1
  else if(nb==2){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.8660254037844385965883021;
          A[0][1] = 0.4330127018922192982941510;
          A[0][2] = 0.09622504486493761721721540;
          A[0][3] = 0.01202813060811720215215193;
          A[1][1] = 0.02405626121623440430430385;
          A[1][2] = 0.01202813060811720215215193;
          A[1][3] = 0.001202813060811720301951366;
          A[2][2] = -0.0004009376869372400825804192;
          A[2][3] = -0.0002004688434686200412902096;
          A[3][3] = -8.591521862940859348054500e-06;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.2500000000000000000000000;
          A[0][2] = 0.1250000000000000000000000;
          A[0][3] = 0.02083333333333333217685102;
          A[1][0] = 0.2500000000000000000000000;
          A[1][1] = 0.1250000000000000000000000;
          A[1][2] = 0.02500000000000000138777878;
          A[1][3] = 0.002083333333333333304421275;
          A[2][1] = -0.000000000000000000000000;
          A[2][2] = -0.000000000000000000000000;
          A[2][3] = -0.0003472222222222222354735821;
          A[3][2] = -2.976190476190476187894757e-05;
          A[3][3] = -1.488095238095238093947378e-05;
        } // |m|=0
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][2] = 0.05379143536399189767793771;
          A[0][3] = 0.02689571768199594883896886;
          A[1][1] = 0.08606629658238704183581547;
          A[1][2] = 0.04303314829119352091790773;
          A[1][3] = 0.002689571768199595057369233;
          A[2][0] = 0.03227486121839513721898385;
          A[2][1] = 0.01613743060919756860949192;
          A[2][2] = 0.001793047845466396632632677;
          A[2][3] = -0.0004482619613665991581581693;
          A[3][1] = -0.0002305347229885367269473784;
          A[3][2] = -0.0001152673614942683634736892;
          A[3][3] = -1.921122691571139278290427e-05;
        } // |m|=0
      } // lb=2
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.08333333333333332870740406;
          A[0][2] = 0.04166666666666666435370203;
          A[0][3] = 0.02083333333333333217685102;
          A[1][0] = -0.2500000000000000000000000;
          A[1][1] = -0.1250000000000000000000000;
          A[1][2] = -0.01666666666666666643537020;
          A[1][3] = 0.002083333333333333304421275;
          A[2][1] = -0.005555555555555555767577314;
          A[2][2] = -0.002777777777777777883788657;
          A[2][3] = -0.0003472222222222222354735821;
          A[3][2] = -2.976190476190476187894757e-05;
          A[3][3] = -1.488095238095238093947378e-05;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][0] = -0.8660254037844385965883021;
          A[0][1] = -0.4330127018922192982941510;
          A[0][2] = 0.000000000000000000000000;
          A[0][3] = 0.03608439182435160819117925;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = -0.04330127018922193260497266;
          A[1][2] = -0.02165063509461096630248633;
          A[1][3] = 0.003608439182435160905854099;
          A[2][0] = -0.02886751345948128724683279;
          A[2][1] = -0.01443375672974064362341640;
          A[2][2] = -0.003608439182435160905854099;
          A[2][3] = -0.0006014065304058601509756832;
          A[3][1] = -0.0002061965247105806108007808;
          A[3][2] = -0.0001030982623552903054003904;
          A[3][3] = -2.577456558882257635009760e-05;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 0.8660254037844385965883021;
          A[0][1] = 0.4330127018922192982941510;
          A[0][2] = 0.07216878364870321638235851;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.01443375672974064362341640;
          A[1][2] = 0.007216878364870321811708198;
          A[2][0] = -0.01443375672974064362341640;
          A[2][1] = -0.007216878364870321811708198;
          A[2][2] = -0.001202813060811720301951366;
          A[3][1] = -0.0001030982623552903054003904;
          A[3][2] = -5.154913117764515270019521e-05;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][1] = -0.3726779962499649490048625;
          A[0][2] = -0.1863389981249824745024313;
          A[0][3] = 0.04658474953124561862560782;
          A[1][0] = -0.2236067977499789916073780;
          A[1][1] = -0.1118033988749894958036890;
          A[1][2] = -0.009316949906249124072066259;
          A[1][3] = 0.004658474953124562036033129;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.003105649968749707879461797;
          A[2][2] = -0.001552824984374853939730898;
          A[2][3] = -0.0007764124921874269698654492;
          A[3][0] = -0.002395787118749774687043175;
          A[3][1] = -0.001197893559374887343521587;
          A[3][2] = -0.0002661985687499750013670918;
          A[3][3] = -3.327482109374687517088648e-05;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.3227486121839513999454141;
          A[0][2] = 0.1613743060919756999727070;
          A[1][0] = 0.1936491673103708510694787;
          A[1][1] = 0.09682458365518542553473935;
          A[1][2] = 0.01613743060919757207893888;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.005379143536399190114738467;
          A[2][2] = -0.002689571768199595057369233;
          A[3][0] = -0.001383208337931220470104487;
          A[3][1] = -0.0006916041689656102350522437;
          A[3][2] = -0.0001152673614942683634736892;
        } // |m|=1
      } // lb=2
    } // la=1
  } // nb=2
  else if(nb==3){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.6324555320336758823529522;
          A[0][1] = 0.3162277660168379411764761;
          A[0][2] = 0.07905694150420948529411902;
          A[0][3] = 0.01317615691736824697077868;
          A[0][4] = 0.001317615691736824783814042;
          A[1][1] = 0.02635231383473649394155736;
          A[1][2] = 0.01317615691736824697077868;
          A[1][3] = 0.002635231383473649567628083;
          A[1][4] = 0.0002196026152894707973023403;
          A[2][2] = 0.000000000000000000000000;
          A[2][3] = 0.000000000000000000000000;
          A[2][4] = -1.568590109210505598212951e-05;
          A[3][3] = -3.137180218421011196425902e-06;
          A[3][4] = -1.568590109210505598212951e-06;
          A[4][4] = -2.489825570175405911022863e-08;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.1825741858350553581402664;
          A[0][2] = 0.09128709291752767907013322;
          A[0][3] = 0.01977887013213099759445512;
          A[0][4] = 0.002282177322938192063489504;
          A[1][0] = 0.1825741858350553581402664;
          A[1][1] = 0.09128709291752767907013322;
          A[1][2] = 0.02434322477800738085407239;
          A[1][3] = 0.004564354645876384126979008;
          A[1][4] = 0.0003803628871563653258448812;
          A[2][1] = 0.001521451548625461303379525;
          A[2][2] = 0.0007607257743127306516897623;
          A[2][3] = 7.245007374406959297415826e-05;
          A[2][4] = -2.716877765402609567124345e-05;
          A[3][2] = -1.086751106161043792968420e-05;
          A[3][3] = -5.433755530805218964842101e-06;
          A[3][4] = -2.716877765402609482421050e-06;
          A[4][3] = -8.625008779055902992750416e-08;
          A[4][4] = -4.312504389527951496375208e-08;
        } // |m|=0
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][2] = 0.03535533905932737586130798;
          A[0][3] = 0.01767766952966368793065399;
          A[0][4] = 0.002946278254943948422123201;
          A[1][1] = 0.05892556509887896670774055;
          A[1][2] = 0.02946278254943948335387027;
          A[1][3] = 0.005892556509887896844246402;
          A[1][4] = 0.0004910463758239913675471278;
          A[2][0] = 0.02357022603955158737698561;
          A[2][1] = 0.01178511301977579368849280;
          A[2][2] = 0.002525381361380527218962744;
          A[2][3] = 0.0002805979290422807659668991;
          A[2][4] = -3.507474113028509574586239e-05;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = -7.014948226057019487985657e-06;
          A[3][4] = -3.507474113028509743992828e-06;
          A[4][2] = -6.680903072435256756061691e-07;
          A[4][3] = -3.340451536217628378030845e-07;
          A[4][4] = -5.567419227029380630051409e-08;
        } // |m|=0
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][3] = 0.006972166887783962790914316;
          A[0][4] = 0.003486083443891981395457158;
          A[1][2] = 0.01394433377556792558182863;
          A[1][3] = 0.006972166887783962790914316;
          A[1][4] = 0.0005810139073153303048563378;
          A[2][1] = 0.008964214570007952159746978;
          A[2][2] = 0.004482107285003976079873489;
          A[2][3] = 0.0006640158940746631590842930;
          A[2][4] = -4.150099337966644744276831e-05;
          A[3][0] = 0.001992047682223989368832662;
          A[3][1] = 0.0009960238411119946844163309;
          A[3][2] = 0.0001494035761667992135044714;
          A[3][3] = -8.300198675933289827366841e-06;
          A[3][4] = -4.150099337966644913683421e-06;
          A[4][1] = -7.904951119936466018235308e-06;
          A[4][2] = -3.952475559968233009117654e-06;
          A[4][3] = -7.904951119936465594718834e-07;
          A[4][4] = -6.587459266613721770095355e-08;
        } // |m|=0
      } // lb=3
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = -0.000000000000000000000000;
          A[0][2] = -0.000000000000000000000000;
          A[0][3] = 0.004564354645876384126979008;
          A[0][4] = 0.002282177322938192063489504;
          A[1][0] = -0.1825741858350553581402664;
          A[1][1] = -0.09128709291752767907013322;
          A[1][2] = -0.01825741858350553650791603;
          A[1][3] = -0.001521451548625461303379525;
          A[1][4] = 0.0003803628871563653258448812;
          A[2][1] = -0.004564354645876384126979008;
          A[2][2] = -0.002282177322938192063489504;
          A[2][3] = -0.0004347004424644175307398952;
          A[2][4] = -2.716877765402609567124345e-05;
          A[3][2] = -3.260253318483131209498671e-05;
          A[3][3] = -1.630126659241565604749336e-05;
          A[3][4] = -2.716877765402609482421050e-06;
          A[4][3] = -8.625008779055902992750416e-08;
          A[4][4] = -4.312504389527951496375208e-08;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][0] = -0.6324555320336758823529522;
          A[0][1] = -0.3162277660168379411764761;
          A[0][2] = -0.04743416490252568701313507;
          A[0][3] = 0.002635231383473649567628083;
          A[0][4] = 0.003952847075210473917761256;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = -0.04743416490252568701313507;
          A[1][2] = -0.02371708245126284350656753;
          A[1][3] = -0.002635231383473649567628083;
          A[1][4] = 0.0006588078458684123919070208;
          A[2][0] = -0.02108185106778919654102467;
          A[2][1] = -0.01054092553389459827051233;
          A[2][2] = -0.003011693009684170748568866;
          A[2][3] = -0.0006274360436842021850750717;
          A[2][4] = -4.705770327631516794638852e-05;
          A[3][1] = -0.0002258769757263128061426649;
          A[3][2] = -0.0001129384878631564030713325;
          A[3][3] = -2.823462196578910076783311e-05;
          A[3][4] = -4.705770327631516794638852e-06;
          A[4][2] = -5.975581368420973657059280e-07;
          A[4][3] = -2.987790684210486828529640e-07;
          A[4][4] = -7.469476710526217071324100e-08;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 0.6324555320336758823529522;
          A[0][1] = 0.3162277660168379411764761;
          A[0][2] = 0.06851601597031488355415974;
          A[0][3] = 0.007905694150420947835522512;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.01581138830084189567104502;
          A[1][2] = 0.007905694150420947835522512;
          A[1][3] = 0.001317615691736824783814042;
          A[2][0] = -0.01054092553389459827051233;
          A[2][1] = -0.005270462766947299135256166;
          A[2][2] = -0.001066641274263143779679752;
          A[2][3] = -9.411540655263033589277705e-05;
          A[3][1] = -0.0001129384878631564030713325;
          A[3][2] = -5.646924393157820153566623e-05;
          A[3][3] = -9.411540655263033589277705e-06;
          A[4][2] = -2.987790684210486828529640e-07;
          A[4][3] = -1.493895342105243414264820e-07;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][1] = -0.2449489742783177992357935;
          A[0][2] = -0.1224744871391588996178967;
          A[0][3] = -0.01020620726159657438991690;
          A[0][4] = 0.005103103630798287194958451;
          A[1][0] = -0.1632993161855451902386704;
          A[1][1] = -0.08164965809277259511933522;
          A[1][2] = -0.02041241452319314877983381;
          A[1][3] = -0.003402069087198858129972301;
          A[1][4] = 0.0008505172717997145324930752;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.004374088826398532005301778;
          A[2][2] = -0.002187044413199266002650889;
          A[2][3] = -0.0004860098695998368834546299;
          A[2][4] = -6.075123369997961043182874e-05;
          A[3][0] = -0.001749635530559412845488798;
          A[3][1] = -0.0008748177652797064227443991;
          A[3][2] = -0.0002187044413199266056860998;
          A[3][3] = -3.645074021998776761434996e-05;
          A[3][4] = -6.075123369997960704369695e-06;
          A[4][1] = -6.942998137140526761289065e-06;
          A[4][2] = -3.471499068570263380644533e-06;
          A[4][3] = -7.714442374600585290321184e-07;
          A[4][4] = -9.643052968250731612901479e-08;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.2121320343559642274122723;
          A[0][2] = 0.1060660171779821137061361;
          A[0][3] = 0.01767766952966368793065399;
          A[1][0] = 0.1414213562373095034452319;
          A[1][1] = 0.07071067811865475172261597;
          A[1][2] = 0.01767766952966368793065399;
          A[1][3] = 0.002946278254943947554761463;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.002525381361380526785281875;
          A[2][2] = -0.001262690680690263392640937;
          A[2][3] = -0.0002104484467817105473701200;
          A[3][0] = -0.001010152544552210670744663;
          A[3][1] = -0.0005050762722761053353723315;
          A[3][2] = -0.0001262690680690263338430829;
          A[3][3] = -2.104484467817105676989108e-05;
          A[4][1] = -4.008541843461153630120541e-06;
          A[4][2] = -2.004270921730576815060270e-06;
          A[4][3] = -3.340451536217627848635253e-07;
        } // |m|=1
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][2] = -0.07245688373094719136879860;
          A[0][3] = -0.03622844186547359568439930;
          A[0][4] = 0.006038073644245600148094955;
          A[1][1] = -0.07245688373094719136879860;
          A[1][2] = -0.03622844186547359568439930;
          A[1][3] = -0.004025382429497066476276057;
          A[1][4] = 0.001006345607374266619069014;
          A[2][0] = -0.02070196678027062808791214;
          A[2][1] = -0.01035098339013531404395607;
          A[2][2] = -0.001725163898355885601712534;
          A[2][3] = 0.000000000000000000000000;
          A[2][4] = -7.188182909816189555384652e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = -0.0003450327796711771203425068;
          A[3][2] = -0.0001725163898355885601712534;
          A[3][3] = -4.312909745889714004281335e-05;
          A[3][4] = -7.188182909816190571824189e-06;
          A[4][0] = -0.0001095342157686276542399167;
          A[4][1] = -5.476710788431382711995837e-05;
          A[4][2] = -1.232259927397061144080381e-05;
          A[4][3] = -1.597373979959153319233217e-06;
          A[4][4] = -1.140981414256538179701511e-07;
        } // |m|=0
        else if(abs(m)==1){
          A[0][2] = 0.05916079783099616185015890;
          A[0][3] = 0.02958039891549808092507945;
          A[1][1] = 0.05916079783099616185015890;
          A[1][2] = 0.02958039891549808092507945;
          A[1][3] = 0.004930066485916346531725996;
          A[2][0] = 0.01690308509457033245282354;
          A[2][1] = 0.008451542547285166226411768;
          A[2][2] = 0.0007042952122737638522009807;
          A[2][3] = -0.0003521476061368819261004903;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = -0.0004225771273642582896365449;
          A[3][2] = -0.0002112885636821291448182725;
          A[3][3] = -3.521476061368818854429089e-05;
          A[4][0] = -6.707573450226321433876447e-05;
          A[4][1] = -3.353786725113160716938224e-05;
          A[4][2] = -6.707573450226322111502805e-06;
          A[4][3] = -5.589644541855267720391548e-07;
        } // |m|=1
      } // lb=3
    } // la=1
  } // nb=3
} // na=1
else if(na==2){
  if(nb==0){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.4082482904638630172300395;
          A[0][1] = 0.2041241452319315086150198;
          A[0][2] = 0.06804138174397716953833992;
          A[1][1] = -0.03402069087198858476916996;
          A[1][2] = -0.01701034543599429238458498;
          A[2][2] = 0.0005670115145331430578021226;
        } // |m|=0
      } // lb=0
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.2357022603955158668309622;
          A[0][2] = 0.1178511301977579334154811;
          A[1][0] = -0.1178511301977579334154811;
          A[1][1] = -0.05892556509887896670774055;
          A[1][2] = -0.02946278254943948335387027;
          A[2][1] = 0.001964185503295965470188511;
          A[2][2] = 0.0009820927516479827350942555;
        } // |m|=0
      } // lb=0
    } // la=1
    else if(la==2){
      if(lb==0){
        if(abs(m)==0){
          A[0][2] = 0.1521451548625461502872724;
          A[1][1] = -0.07607257743127307514363622;
          A[1][2] = -0.03803628871563653757181811;
          A[2][0] = 0.01521451548625461433483785;
          A[2][1] = 0.007607257743127307167418927;
          A[2][2] = 0.001267876290521218005796777;
        } // |m|=0
      } // lb=0
    } // la=2
  } // nb=0
  else if(nb==1){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.8660254037844385965883021;
          A[0][1] = 0.4330127018922192982941510;
          A[0][2] = 0.09622504486493761721721540;
          A[0][3] = 0.01202813060811720215215193;
          A[1][1] = -0.02405626121623440430430385;
          A[1][2] = -0.01202813060811720215215193;
          A[1][3] = -0.001202813060811720301951366;
          A[2][2] = -0.0004009376869372400825804192;
          A[2][3] = -0.0002004688434686200412902096;
          A[3][3] = 8.591521862940859348054500e-06;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.08333333333333332870740406;
          A[0][2] = 0.04166666666666666435370203;
          A[0][3] = 0.02083333333333333217685102;
          A[1][0] = 0.2500000000000000000000000;
          A[1][1] = 0.1250000000000000000000000;
          A[1][2] = 0.01666666666666666643537020;
          A[1][3] = -0.002083333333333333304421275;
          A[2][1] = -0.005555555555555555767577314;
          A[2][2] = -0.002777777777777777883788657;
          A[2][3] = -0.0003472222222222222354735821;
          A[3][2] = 2.976190476190476187894757e-05;
          A[3][3] = 1.488095238095238093947378e-05;
        } // |m|=0
      } // lb=1
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.2500000000000000000000000;
          A[0][2] = 0.1250000000000000000000000;
          A[0][3] = 0.02083333333333333217685102;
          A[1][0] = -0.2500000000000000000000000;
          A[1][1] = -0.1250000000000000000000000;
          A[1][2] = -0.02500000000000000138777878;
          A[1][3] = -0.002083333333333333304421275;
          A[2][1] = -0.000000000000000000000000;
          A[2][2] = -0.000000000000000000000000;
          A[2][3] = -0.0003472222222222222354735821;
          A[3][2] = 2.976190476190476187894757e-05;
          A[3][3] = 1.488095238095238093947378e-05;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][0] = -0.8660254037844385965883021;
          A[0][1] = -0.4330127018922192982941510;
          A[0][2] = 0.000000000000000000000000;
          A[0][3] = 0.03608439182435160819117925;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.04330127018922193260497266;
          A[1][2] = 0.02165063509461096630248633;
          A[1][3] = -0.003608439182435160905854099;
          A[2][0] = -0.02886751345948128724683279;
          A[2][1] = -0.01443375672974064362341640;
          A[2][2] = -0.003608439182435160905854099;
          A[2][3] = -0.0006014065304058601509756832;
          A[3][1] = 0.0002061965247105806108007808;
          A[3][2] = 0.0001030982623552903054003904;
          A[3][3] = 2.577456558882257635009760e-05;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 0.8660254037844385965883021;
          A[0][1] = 0.4330127018922192982941510;
          A[0][2] = 0.07216878364870321638235851;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = -0.01443375672974064362341640;
          A[1][2] = -0.007216878364870321811708198;
          A[2][0] = -0.01443375672974064362341640;
          A[2][1] = -0.007216878364870321811708198;
          A[2][2] = -0.001202813060811720301951366;
          A[3][1] = 0.0001030982623552903054003904;
          A[3][2] = 5.154913117764515270019521e-05;
        } // |m|=1
      } // lb=1
    } // la=1
    else if(la==2){
      if(lb==0){
        if(abs(m)==0){
          A[0][2] = 0.05379143536399189767793771;
          A[0][3] = 0.02689571768199594883896886;
          A[1][1] = -0.08606629658238704183581547;
          A[1][2] = -0.04303314829119352091790773;
          A[1][3] = -0.002689571768199595057369233;
          A[2][0] = 0.03227486121839513721898385;
          A[2][1] = 0.01613743060919756860949192;
          A[2][2] = 0.001793047845466396632632677;
          A[2][3] = -0.0004482619613665991581581693;
          A[3][1] = 0.0002305347229885367269473784;
          A[3][2] = 0.0001152673614942683634736892;
          A[3][3] = 1.921122691571139278290427e-05;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = -0.3726779962499649490048625;
          A[0][2] = -0.1863389981249824745024313;
          A[0][3] = 0.04658474953124561862560782;
          A[1][0] = 0.2236067977499789916073780;
          A[1][1] = 0.1118033988749894958036890;
          A[1][2] = 0.009316949906249124072066259;
          A[1][3] = -0.004658474953124562036033129;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.003105649968749707879461797;
          A[2][2] = -0.001552824984374853939730898;
          A[2][3] = -0.0007764124921874269698654492;
          A[3][0] = 0.002395787118749774687043175;
          A[3][1] = 0.001197893559374887343521587;
          A[3][2] = 0.0002661985687499750013670918;
          A[3][3] = 3.327482109374687517088648e-05;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.3227486121839513999454141;
          A[0][2] = 0.1613743060919756999727070;
          A[1][0] = -0.1936491673103708510694787;
          A[1][1] = -0.09682458365518542553473935;
          A[1][2] = -0.01613743060919757207893888;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.005379143536399190114738467;
          A[2][2] = -0.002689571768199595057369233;
          A[3][0] = 0.001383208337931220470104487;
          A[3][1] = 0.0006916041689656102350522437;
          A[3][2] = 0.0001152673614942683634736892;
        } // |m|=1
      } // lb=1
    } // la=2
  } // nb=1
  else if(nb==2){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.1111111111111111049432054;
          A[0][3] = 0.01388888888888888811790068;
          A[0][4] = 0.001388888888888888725053894;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[2][2] = -0.0004629629629629629264213342;
          A[2][3] = -0.0002314814814814814632106671;
          A[2][4] = -3.306878306878306423687713e-05;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[4][4] = 3.936759889140841087720195e-08;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.1443375672974064327647170;
          A[0][2] = 0.07216878364870321638235851;
          A[0][3] = 0.01683938285136408335995739;
          A[0][4] = 0.002405626121623440603902733;
          A[1][0] = 0.2886751345948128655294340;
          A[1][1] = 0.1443375672974064327647170;
          A[1][2] = 0.02646188733785784577556832;
          A[1][3] = 0.001202813060811720301951366;
          A[1][4] = -0.000000000000000000000000;
          A[2][1] = -0.002405626121623440603902733;
          A[2][2] = -0.001202813060811720301951366;
          A[2][3] = -0.0003150224683078315128167968;
          A[2][4] = -5.727681241960572898703000e-05;
          A[3][2] = -1.718304372588171869610900e-05;
          A[3][3] = -8.591521862940859348054500e-06;
          A[3][4] = -0.000000000000000000000000;
          A[4][3] = 1.363733629038231604734362e-07;
          A[4][4] = 6.818668145191158023671812e-08;
        } // |m|=0
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][2] = 0.006211299937499415758923593;
          A[0][3] = 0.003105649968749707879461797;
          A[0][4] = 0.003105649968749707879461797;
          A[1][1] = 0.05590169943749474096295060;
          A[1][2] = 0.02795084971874737048147530;
          A[1][3] = 0.004658474953124562036033129;
          A[1][4] = 0.000000000000000000000000;
          A[2][0] = 0.03726779962499649628826504;
          A[2][1] = 0.01863389981249824814413252;
          A[2][2] = 0.002735929734374742666042080;
          A[2][3] = -0.0001848601171874826067098585;
          A[2][4] = -7.394404687499304539444883e-05;
          A[3][1] = -0.0003992978531249624478405291;
          A[3][2] = -0.0001996489265624812239202646;
          A[3][3] = -3.327482109374686839462290e-05;
          A[3][4] = 0.000000000000000000000000;
          A[4][2] = 1.056343526785614879754294e-06;
          A[4][3] = 5.281717633928074398771469e-07;
          A[4][4] = 8.802862723213457772448775e-08;
        } // |m|=0
      } // lb=2
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.1443375672974064327647170;
          A[0][2] = 0.07216878364870321638235851;
          A[0][3] = 0.01683938285136408335995739;
          A[0][4] = 0.002405626121623440603902733;
          A[1][0] = -0.2886751345948128655294340;
          A[1][1] = -0.1443375672974064327647170;
          A[1][2] = -0.02646188733785784577556832;
          A[1][3] = -0.001202813060811720301951366;
          A[1][4] = -0.000000000000000000000000;
          A[2][1] = -0.002405626121623440603902733;
          A[2][2] = -0.001202813060811720301951366;
          A[2][3] = -0.0003150224683078315128167968;
          A[2][4] = -5.727681241960572898703000e-05;
          A[3][2] = 1.718304372588171869610900e-05;
          A[3][3] = 8.591521862940859348054500e-06;
          A[3][4] = -0.000000000000000000000000;
          A[4][3] = 1.363733629038231604734362e-07;
          A[4][4] = 6.818668145191158023671812e-08;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][0] = -1.000000000000000000000000;
          A[0][1] = -0.5000000000000000000000000;
          A[0][2] = -0.05000000000000000277555756;
          A[0][3] = 0.01666666666666666643537020;
          A[0][4] = 0.004166666666666666608842551;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[2][0] = -0.03333333333333333287074041;
          A[2][1] = -0.01666666666666666643537020;
          A[2][2] = -0.003571428571428571317053491;
          A[2][3] = -0.0003968253968253968250526342;
          A[2][4] = -9.920634920634920626315856e-05;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[4][2] = 9.448223733938019139924061e-07;
          A[4][3] = 4.724111866969009569962030e-07;
          A[4][4] = 1.181027966742252392490508e-07;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.1000000000000000055511151;
          A[0][3] = 0.008333333333333333217685102;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[2][0] = -0.01666666666666666643537020;
          A[2][1] = -0.008333333333333333217685102;
          A[2][2] = -0.001785714285714285658526745;
          A[2][3] = -0.0001984126984126984125263171;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[4][2] = 4.724111866969009569962030e-07;
          A[4][3] = 2.362055933484504784981015e-07;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][1] = -0.2581988897471610977518708;
          A[0][2] = -0.1290994448735805488759354;
          A[0][3] = -0.01075828707279838022947693;
          A[0][4] = 0.005379143536399190114738467;
          A[1][0] = -0.2581988897471610977518708;
          A[1][1] = -0.1290994448735805488759354;
          A[1][2] = -0.01075828707279838022947693;
          A[1][3] = 0.005379143536399190114738467;
          A[1][4] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.001536898153256911368422233;
          A[2][2] = 0.0007684490766284556842111164;
          A[2][3] = -0.0001280748461047426321052223;
          A[2][4] = -0.0001280748461047426321052223;
          A[3][0] = -0.002766416675862440506528106;
          A[3][1] = -0.001383208337931220253264053;
          A[3][2] = -0.0003073796306513822845264683;
          A[3][3] = -3.842245383142278556580854e-05;
          A[3][4] = -0.000000000000000000000000;
          A[4][1] = 1.097784395183508134822160e-05;
          A[4][2] = 5.488921975917540674110798e-06;
          A[4][3] = 1.219760439092786863526452e-06;
          A[4][4] = 1.524700548865983579408065e-07;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.2236067977499789916073780;
          A[0][2] = 0.1118033988749894958036890;
          A[0][3] = 0.01863389981249824814413252;
          A[1][0] = 0.2236067977499789916073780;
          A[1][1] = 0.1118033988749894958036890;
          A[1][2] = 0.01863389981249824814413252;
          A[1][3] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.005323971374999499593660968;
          A[2][2] = -0.002661985687499749796830484;
          A[2][3] = -0.0004436642812499582994717473;
          A[3][0] = -0.001597191412499849791362116;
          A[3][1] = -0.0007985957062499248956810582;
          A[3][2] = -0.0001330992843749875006835459;
          A[3][3] = -0.000000000000000000000000;
          A[4][1] = 6.338061160713689702042237e-06;
          A[4][2] = 3.169030580356844851021118e-06;
          A[4][3] = 5.281717633928075457562653e-07;
        } // |m|=1
      } // lb=2
    } // la=1
    else if(la==2){
      if(lb==0){
        if(abs(m)==0){
          A[0][2] = 0.006211299937499415758923593;
          A[0][3] = 0.003105649968749707879461797;
          A[0][4] = 0.003105649968749707879461797;
          A[1][1] = -0.05590169943749474096295060;
          A[1][2] = -0.02795084971874737048147530;
          A[1][3] = -0.004658474953124562036033129;
          A[1][4] = 0.000000000000000000000000;
          A[2][0] = 0.03726779962499649628826504;
          A[2][1] = 0.01863389981249824814413252;
          A[2][2] = 0.002735929734374742666042080;
          A[2][3] = -0.0001848601171874826067098585;
          A[2][4] = -7.394404687499304539444883e-05;
          A[3][1] = 0.0003992978531249624478405291;
          A[3][2] = 0.0001996489265624812239202646;
          A[3][3] = 3.327482109374686839462290e-05;
          A[3][4] = 0.000000000000000000000000;
          A[4][2] = 1.056343526785614879754294e-06;
          A[4][3] = 5.281717633928074398771469e-07;
          A[4][4] = 8.802862723213457772448775e-08;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = -0.2581988897471610977518708;
          A[0][2] = -0.1290994448735805488759354;
          A[0][3] = -0.01075828707279838022947693;
          A[0][4] = 0.005379143536399190114738467;
          A[1][0] = 0.2581988897471610977518708;
          A[1][1] = 0.1290994448735805488759354;
          A[1][2] = 0.01075828707279838022947693;
          A[1][3] = -0.005379143536399190114738467;
          A[1][4] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.001536898153256911368422233;
          A[2][2] = 0.0007684490766284556842111164;
          A[2][3] = -0.0001280748461047426321052223;
          A[2][4] = -0.0001280748461047426321052223;
          A[3][0] = 0.002766416675862440506528106;
          A[3][1] = 0.001383208337931220253264053;
          A[3][2] = 0.0003073796306513822845264683;
          A[3][3] = 3.842245383142278556580854e-05;
          A[3][4] = -0.000000000000000000000000;
          A[4][1] = 1.097784395183508134822160e-05;
          A[4][2] = 5.488921975917540674110798e-06;
          A[4][3] = 1.219760439092786863526452e-06;
          A[4][4] = 1.524700548865983579408065e-07;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.2236067977499789916073780;
          A[0][2] = 0.1118033988749894958036890;
          A[0][3] = 0.01863389981249824814413252;
          A[1][0] = -0.2236067977499789916073780;
          A[1][1] = -0.1118033988749894958036890;
          A[1][2] = -0.01863389981249824814413252;
          A[1][3] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.005323971374999499593660968;
          A[2][2] = -0.002661985687499749796830484;
          A[2][3] = -0.0004436642812499582994717473;
          A[3][0] = 0.001597191412499849791362116;
          A[3][1] = 0.0007985957062499248956810582;
          A[3][2] = 0.0001330992843749875006835459;
          A[3][3] = -0.000000000000000000000000;
          A[4][1] = 6.338061160713689702042237e-06;
          A[4][2] = 3.169030580356844851021118e-06;
          A[4][3] = 5.281717633928075457562653e-07;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = -0.02777777777777777970524831;
          A[0][3] = -0.05555555555555555941049661;
          A[0][4] = 0.006944444444444444926312077;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[2][0] = 0.02380952380952381167156240;
          A[2][1] = 0.01190476190476190583578120;
          A[2][2] = 0.002645502645502645789471474;
          A[2][3] = 0.0003306878306878307236839343;
          A[2][4] = -0.0001653439153439153618419671;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[4][0] = 0.0001700680272108843690825886;
          A[4][1] = 8.503401360544218454129428e-05;
          A[4][2] = 1.889644746787603912688107e-05;
          A[4][3] = 2.362055933484504890860134e-06;
          A[4][4] = 1.968379944570420742383445e-07;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = -1.000000000000000000000000;
          A[0][1] = -0.5000000000000000000000000;
          A[0][2] = 0.000000000000000000000000;
          A[0][3] = 0.04166666666666666435370203;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[2][0] = -0.01190476190476190410105772;
          A[2][1] = -0.005952380952380952050528862;
          A[2][2] = -0.002976190476190476025264431;
          A[2][3] = -0.0009920634920634920084214770;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[4][0] = 0.0001133786848072562415375500;
          A[4][1] = 5.668934240362812076877500e-05;
          A[4][2] = 1.181027966742252487781714e-05;
          A[4][3] = 1.181027966742252445430067e-06;
        } // |m|=1
        else if(abs(m)==2){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.08333333333333332870740406;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[2][0] = -0.02380952380952380820211545;
          A[2][1] = -0.01190476190476190410105772;
          A[2][2] = -0.001984126984126984016842954;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[4][0] = 2.834467120181406038438750e-05;
          A[4][1] = 1.417233560090703019219375e-05;
          A[4][2] = 2.362055933484504890860134e-06;
        } // |m|=2
      } // lb=2
    } // la=2
  } // nb=2
  else if(nb==3){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.9128709291752767907013322;
          A[0][1] = 0.4564354645876383953506661;
          A[0][2] = 0.1065016084037822968744180;
          A[0][3] = 0.01521451548625461260011438;
          A[0][4] = 0.001521451548625461303379525;
          A[0][5] = 0.0001267876290521217843166452;
          A[1][1] = 0.01521451548625461260011438;
          A[1][2] = 0.007607257743127306300057189;
          A[1][3] = 0.001521451548625461303379525;
          A[1][4] = 0.0001267876290521217843166452;
          A[1][5] = 9.056259218008699121769782e-06;
          A[2][2] = -0.0002535752581042435686332903;
          A[2][3] = -0.0001267876290521217843166452;
          A[2][4] = -2.716877765402609567124345e-05;
          A[2][5] = -3.018753072669566232751103e-06;
          A[3][3] = -1.811251843601739654947367e-06;
          A[3][4] = -9.056259218008698274736835e-07;
          A[3][5] = -1.006251024223188638371249e-07;
          A[4][4] = 7.187507315879919436352217e-09;
          A[4][5] = 3.593753657939959718176109e-09;
          A[5][5] = 9.075135499848382669773610e-11;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.1581138830084189705882380;
          A[0][2] = 0.07905694150420948529411902;
          A[0][3] = 0.01756820922315766378418722;
          A[0][4] = 0.002196026152894707973023403;
          A[0][5] = 0.0002196026152894707973023403;
          A[1][0] = 0.2635231383473649602322553;
          A[1][1] = 0.1317615691736824801161276;
          A[1][2] = 0.02898754521821014437654718;
          A[1][3] = 0.003513641844631532756837444;
          A[1][4] = 0.0002509744174736808957140721;
          A[1][5] = 1.568590109210505598212951e-05;
          A[2][1] = -0.000000000000000000000000;
          A[2][2] = -0.000000000000000000000000;
          A[2][3] = -7.320087176315693243411342e-05;
          A[2][4] = -3.660043588157846621705671e-05;
          A[2][5] = -5.228633697368352558731801e-06;
          A[3][2] = -1.882308131052606717855541e-05;
          A[3][3] = -9.411540655263033589277705e-06;
          A[3][4] = -1.917165689035062491930601e-06;
          A[3][5] = -1.742877899122783939192657e-07;
          A[4][3] = -0.000000000000000000000000;
          A[4][4] = -0.000000000000000000000000;
          A[4][5] = 6.224563925438514777557158e-09;
          A[5][4] = 3.143719154261876171169792e-10;
          A[5][5] = 1.571859577130938085584896e-10;
        } // |m|=0
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][2] = 0.01360827634879543425461268;
          A[0][3] = 0.006804138174397717127306340;
          A[0][4] = 0.001701034543599429281826585;
          A[0][5] = 0.0002835057572665715289010613;
          A[1][1] = 0.05443310539518173701845072;
          A[1][2] = 0.02721655269759086850922536;
          A[1][3] = 0.005346108565598205880631255;
          A[1][4] = 0.0004050082246665307633172459;
          A[1][5] = 2.025041123332654019874137e-05;
          A[2][0] = 0.03402069087198858476916996;
          A[2][1] = 0.01701034543599429238458498;
          A[2][2] = 0.003321067442265552335095569;
          A[2][3] = 0.0002430049347999184688323693;
          A[2][4] = -2.025041123332654019874137e-05;
          A[2][5] = -6.750137077775512552547509e-06;
          A[3][1] = -0.0001458029608799510704573998;
          A[3][2] = -7.290148043997553522869992e-05;
          A[3][3] = -1.890038381777143480831985e-05;
          A[3][4] = -3.375068538887756276273755e-06;
          A[3][5] = -2.250045692591837641041475e-07;
          A[4][2] = -5.785831780950440026532072e-07;
          A[4][3] = -2.892915890475220013266036e-07;
          A[4][4] = -3.214350989416911199378317e-08;
          A[4][5] = 8.035877473542277998445791e-09;
          A[5][3] = 2.435114385921902173110660e-09;
          A[5][4] = 1.217557192960951086555330e-09;
          A[5][5] = 2.029261988268251897090197e-10;
        } // |m|=0
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][3] = -0.000000000000000000000000;
          A[0][4] = -0.000000000000000000000000;
          A[0][5] = 0.0003354485357914222063563381;
          A[1][2] = 0.006900655593423542406850135;
          A[1][3] = 0.003450327796711771203425068;
          A[1][4] = 0.0006229758521840697343616156;
          A[1][5] = 2.396060969938729851794884e-05;
          A[2][1] = 0.008625819491779427791722235;
          A[2][2] = 0.004312909745889713895861117;
          A[2][3] = 0.0007667395103803935525743629;
          A[2][4] = 2.396060969938729851794884e-05;
          A[2][5] = -7.986869899795766172649614e-06;
          A[3][0] = 0.002875273163926475930574078;
          A[3][1] = 0.001437636581963237965287039;
          A[3][2] = 0.0002268271051541997484612273;
          A[3][3] = -6.389495919836613276932870e-06;
          A[3][4] = -5.590808929857035982041551e-06;
          A[3][5] = -2.662289966598588688923498e-07;
          A[4][1] = -1.825570262810460791060886e-05;
          A[4][2] = -9.127851314052303955304431e-06;
          A[4][3] = -1.711472121384807097498699e-06;
          A[4][4] = -9.508178452137817502434769e-08;
          A[4][5] = 9.508178452137816509818034e-09;
          A[5][2] = 2.881266197617520264781062e-08;
          A[5][3] = 1.440633098808760132390531e-08;
          A[5][4] = 2.881266197617520430217184e-09;
          A[5][5] = 2.401055164681266680522398e-10;
        } // |m|=0
      } // lb=3
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.05270462766947298788311471;
          A[0][2] = 0.02635231383473649394155736;
          A[0][3] = 0.007905694150420947835522512;
          A[0][4] = 0.001756820922315766378418722;
          A[0][5] = 0.0002196026152894707973023403;
          A[1][0] = -0.2635231383473649602322553;
          A[1][1] = -0.1317615691736824801161276;
          A[1][2] = -0.02635231383473649394155736;
          A[1][3] = -0.002196026152894707973023403;
          A[1][4] = 3.137180218421011196425902e-05;
          A[1][5] = 1.568590109210505598212951e-05;
          A[2][1] = -0.003513641844631532756837444;
          A[2][2] = -0.001756820922315766378418722;
          A[2][3] = -0.0003764616262105213435711082;
          A[2][4] = -4.182906957894682046985441e-05;
          A[2][5] = -5.228633697368352558731801e-06;
          A[3][2] = -0.000000000000000000000000;
          A[3][3] = -0.000000000000000000000000;
          A[3][4] = -3.485755798245567878385315e-07;
          A[3][5] = -1.742877899122783939192657e-07;
          A[4][3] = 9.959302280701623644091453e-08;
          A[4][4] = 4.979651140350811822045727e-08;
          A[4][5] = 6.224563925438514777557158e-09;
          A[5][4] = 3.143719154261876171169792e-10;
          A[5][5] = 1.571859577130938085584896e-10;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][0] = -0.9128709291752767907013322;
          A[0][1] = -0.4564354645876383953506661;
          A[0][2] = -0.07302967433402214603166414;
          A[0][3] = 0.001521451548625461303379525;
          A[0][4] = 0.002282177322938192063489504;
          A[0][5] = 0.0003803628871563653258448812;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = -0.02738612787525830302715057;
          A[1][2] = -0.01369306393762915151357529;
          A[1][3] = -0.002064827101705983027069014;
          A[1][4] = 0.0001086751106161043826849738;
          A[1][5] = 2.716877765402609567124345e-05;
          A[2][0] = -0.03042903097250922520022876;
          A[2][1] = -0.01521451548625461260011438;
          A[2][2] = -0.003477603539715340245919162;
          A[2][3] = -0.0004709254793364523340032379;
          A[2][4] = -5.433755530805219134248690e-05;
          A[2][5] = -9.056259218008699121769782e-06;
          A[3][1] = -0.0001304101327393252483799468;
          A[3][2] = -6.520506636966262418997342e-05;
          A[3][3] = -1.328251351974609023825873e-05;
          A[3][4] = -1.207501229067826366045499e-06;
          A[3][5] = -3.018753072669565915113748e-07;
          A[4][2] = 5.175005267433542325045842e-07;
          A[4][3] = 2.587502633716771162522921e-07;
          A[4][4] = 6.468756584291927906307302e-08;
          A[4][5] = 1.078126097381987874093802e-08;
          A[5][3] = 2.178032519963611737348090e-09;
          A[5][4] = 1.089016259981805868674045e-09;
          A[5][5] = 2.722540649954514671685112e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 0.9128709291752767907013322;
          A[0][1] = 0.4564354645876383953506661;
          A[0][2] = 0.1004158022092804525282617;
          A[0][3] = 0.01217161238900369042703620;
          A[0][4] = 0.0007607257743127306516897623;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.009128709291752768253958017;
          A[1][2] = 0.004564354645876384126979008;
          A[1][3] = 0.0008694008849288350614797904;
          A[1][4] = 5.433755530805219134248690e-05;
          A[2][0] = -0.01521451548625461260011438;
          A[2][1] = -0.007607257743127306300057189;
          A[2][2] = -0.001738801769857670122959581;
          A[2][3] = -0.0002354627396682261670016190;
          A[2][4] = -1.811251843601739824353956e-05;
          A[3][1] = -6.520506636966262418997342e-05;
          A[3][2] = -3.260253318483131209498671e-05;
          A[3][3] = -6.641256759873045119129363e-06;
          A[3][4] = -6.037506145339131830227495e-07;
          A[4][2] = 2.587502633716771162522921e-07;
          A[4][3] = 1.293751316858385581261460e-07;
          A[4][4] = 2.156252194763975748187604e-08;
          A[5][3] = 1.089016259981805868674045e-09;
          A[5][4] = 5.445081299909029343370225e-10;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][1] = -0.2357022603955158668309622;
          A[0][2] = -0.1178511301977579334154811;
          A[0][3] = -0.01767766952966368793065399;
          A[0][4] = 0.0009820927516479827350942555;
          A[0][5] = 0.0004910463758239913675471278;
          A[1][0] = -0.2357022603955158668309622;
          A[1][1] = -0.1178511301977579334154811;
          A[1][2] = -0.02104484467817105755593765;
          A[1][3] = -0.0007014948226057019691273564;
          A[1][4] = 0.0002805979290422807659668991;
          A[1][5] = 3.507474113028509574586239e-05;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.001964185503295965470188511;
          A[2][2] = -0.0009820927516479827350942555;
          A[2][3] = -0.0002104484467817105744751743;
          A[2][4] = -2.338316075352339716390826e-05;
          A[2][5] = -1.169158037676169858195413e-05;
          A[3][0] = -0.002525381361380527218962744;
          A[3][1] = -0.001262690680690263609481372;
          A[3][2] = -0.0002805979290422807659668991;
          A[3][3] = -3.507474113028509574586239e-05;
          A[3][4] = -3.117754767136453152828789e-06;
          A[3][5] = -3.897193458920566441035986e-07;
          A[4][1] = -0.000000000000000000000000;
          A[4][2] = -0.000000000000000000000000;
          A[4][3] = 5.567419227029380630051409e-08;
          A[4][4] = 2.783709613514690315025704e-08;
          A[4][5] = 1.391854806757345157512852e-08;
          A[5][2] = 2.530645103195173043739026e-08;
          A[5][3] = 1.265322551597586521869513e-08;
          A[5][4] = 2.811827892439081113755551e-09;
          A[5][5] = 3.514784865548851392194438e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.2041241452319314808594442;
          A[0][2] = 0.1020620726159657404297221;
          A[0][3] = 0.02041241452319314877983381;
          A[0][4] = 0.001701034543599429064986150;
          A[1][0] = 0.2041241452319314808594442;
          A[1][1] = 0.1020620726159657404297221;
          A[1][2] = 0.02187044413199265915914715;
          A[1][3] = 0.002430049347999184254642824;
          A[1][4] = 0.0001215024673999592208636575;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.003402069087198858129972301;
          A[2][2] = -0.001701034543599429064986150;
          A[2][3] = -0.0003645074021998776490384453;
          A[2][4] = -4.050082246665307362121916e-05;
          A[3][0] = -0.001458029608799510596153781;
          A[3][1] = -0.0007290148043997552980768906;
          A[3][2] = -0.0001620032898666122944848766;
          A[3][3] = -2.025041123332653681060958e-05;
          A[3][4] = -1.350027415555102425806207e-06;
          A[4][1] = -0.000000000000000000000000;
          A[4][2] = -0.000000000000000000000000;
          A[4][3] = 9.643052968250731612901479e-08;
          A[4][4] = 4.821526484125365806450740e-08;
          A[5][2] = 1.461068631553141138430274e-08;
          A[5][3] = 7.305343157765705692151368e-09;
          A[5][4] = 1.217557192960950879760177e-09;
        } // |m|=1
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][2] = -0.04183300132670377674548590;
          A[0][3] = -0.02091650066335188837274295;
          A[0][4] = -0.002324055629261321219425351;
          A[0][5] = 0.0005810139073153303048563378;
          A[1][1] = -0.07171371656006361727797582;
          A[1][2] = -0.03585685828003180863898791;
          A[1][3] = -0.004814115232041307930466179;
          A[1][4] = 0.0005810139073153303048563378;
          A[1][5] = 4.150099337966644744276831e-05;
          A[2][0] = -0.02988071523335984053248993;
          A[2][1] = -0.01494035761667992026624496;
          A[2][2] = -0.001992047682223989368832662;
          A[2][3] = 0.0002490059602779986711040827;
          A[2][4] = 5.533465783955525873951536e-05;
          A[2][5] = -1.383366445988881468487884e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = -0.0001438701103828436700122345;
          A[3][2] = -7.193505519142183500611726e-05;
          A[3][3] = -2.490059602779986778803463e-05;
          A[3][4] = -6.455710081281447361163228e-06;
          A[3][5] = -4.611221486629605106717851e-07;
          A[4][0] = -0.0001580990223987293068121790;
          A[4][1] = -7.904951119936465340608950e-05;
          A[4][2] = -1.699564490786340176979932e-05;
          A[4][3] = -1.910363187317979485381220e-06;
          A[4][4] = -6.587459266613721770095355e-08;
          A[4][5] = 1.646864816653430442523839e-08;
          A[5][1] = 3.992399555523467595071114e-07;
          A[5][2] = 1.996199777761733797535557e-07;
          A[5][3] = 4.491449499963900845931657e-08;
          A[5][4] = 5.822249351805057019769457e-09;
          A[5][5] = 4.158749537003612009267360e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][2] = 0.03415650255319865880032637;
          A[0][3] = 0.01707825127659932940016319;
          A[0][4] = 0.002846375212766554755466908;
          A[1][1] = 0.05855400437691198750611576;
          A[1][2] = 0.02927700218845599375305788;
          A[1][3] = 0.005286125395137887712782021;
          A[1][4] = 0.0002033125151976110617062232;
          A[2][0] = 0.02439750182371332870578939;
          A[2][1] = 0.01219875091185666435289470;
          A[2][2] = 0.001626500121580888493649786;
          A[2][3] = -0.0002033125151976110617062232;
          A[2][4] = -6.777083839920368271789869e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = -0.0005150583718339480591291712;
          A[3][2] = -0.0002575291859169740295645856;
          A[3][3] = -4.743958687944258332353994e-05;
          A[3][4] = -2.259027946640122954904311e-06;
          A[4][0] = -9.681548342743384019844222e-05;
          A[4][1] = -4.840774171371692009922111e-05;
          A[4][2] = -8.713393508469046227723522e-06;
          A[4][3] = -3.227182780914461212893132e-07;
          A[4][4] = 8.067956952286153032232829e-08;
          A[5][1] = 2.444835440086713088197426e-07;
          A[5][2] = 1.222417720043356544098713e-07;
          A[5][3] = 2.444835440086713220546324e-08;
          A[5][4] = 2.037362866738927545925168e-09;
        } // |m|=1
      } // lb=3
    } // la=1
    else if(la==2){
      if(lb==0){
        if(abs(m)==0){
          A[0][2] = -0.006804138174397717127306340;
          A[0][3] = -0.003402069087198858563653170;
          A[0][4] = 0.000000000000000000000000;
          A[0][5] = 0.0002835057572665715289010613;
          A[1][1] = -0.02721655269759086850922536;
          A[1][2] = -0.01360827634879543425461268;
          A[1][3] = -0.003159064152398939877980366;
          A[1][4] = -0.0004455090471331838233859379;
          A[1][5] = 2.025041123332654019874137e-05;
          A[2][0] = 0.03402069087198858476916996;
          A[2][1] = 0.01701034543599429238458498;
          A[2][2] = 0.003078062507465633649422765;
          A[2][3] = 0.0001215024673999592344161846;
          A[2][4] = -4.050082246665308039748274e-05;
          A[2][5] = -6.750137077775512552547509e-06;
          A[3][1] = 0.0004374088826398532113721995;
          A[3][2] = 0.0002187044413199266056860998;
          A[3][3] = 4.185084988220818239977247e-05;
          A[3][4] = 2.700054831110205275128888e-06;
          A[3][5] = -2.250045692591837641041475e-07;
          A[4][2] = 1.735749534285131902080503e-06;
          A[4][3] = 8.678747671425659510402516e-07;
          A[4][4] = 1.607175494708455533514709e-07;
          A[4][5] = 8.035877473542277998445791e-09;
          A[5][3] = 2.435114385921902173110660e-09;
          A[5][4] = 1.217557192960951086555330e-09;
          A[5][5] = 2.029261988268251897090197e-10;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = -0.1414213562373095034452319;
          A[0][2] = -0.07071067811865475172261597;
          A[0][3] = -0.01374929852307175785763871;
          A[0][4] = -0.0009820927516479827350942555;
          A[0][5] = 0.0004910463758239913675471278;
          A[1][0] = 0.2357022603955158668309622;
          A[1][1] = 0.1178511301977579334154811;
          A[1][2] = 0.01431049438115632004009381;
          A[1][3] = -0.002665680325901667330895650;
          A[1][4] = -0.0007014948226057019691273564;
          A[1][5] = 3.507474113028509574586239e-05;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.004208968935634212031604573;
          A[2][2] = 0.002104484467817106015802286;
          A[2][3] = 0.0002572147682887573959080452;
          A[2][4] = -4.676632150704679432781652e-05;
          A[2][5] = -1.169158037676169858195413e-05;
          A[3][0] = 0.002525381361380527218962744;
          A[3][1] = 0.001262690680690263609481372;
          A[3][2] = 0.0003086577219465088642476325;
          A[3][3] = 4.910463758239913810996549e-05;
          A[3][4] = 3.897193458920566758673341e-06;
          A[3][5] = -3.897193458920566441035986e-07;
          A[4][1] = 1.603416737384461452048216e-05;
          A[4][2] = 8.017083686922307260241082e-06;
          A[4][3] = 1.837248344919695660856524e-06;
          A[4][4] = 2.505338652163221018825338e-07;
          A[4][5] = 1.391854806757345157512852e-08;
          A[5][2] = 2.530645103195173043739026e-08;
          A[5][3] = 1.265322551597586521869513e-08;
          A[5][4] = 2.811827892439081113755551e-09;
          A[5][5] = 3.514784865548851392194438e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.1224744871391588996178967;
          A[0][2] = 0.06123724356957944980894837;
          A[0][3] = 0.01360827634879543251988920;
          A[0][4] = 0.001701034543599429064986150;
          A[1][0] = -0.2041241452319314808594442;
          A[1][1] = -0.1020620726159657404297221;
          A[1][2] = -0.01895438491439363840052046;
          A[1][3] = -0.0009720197391996737669092599;
          A[1][4] = 0.0001215024673999592208636575;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.004374088826398532005301778;
          A[2][2] = -0.002187044413199266002650889;
          A[2][3] = -0.0004455090471331838233859379;
          A[2][4] = -4.050082246665307362121916e-05;
          A[3][0] = 0.001458029608799510596153781;
          A[3][1] = 0.0007290148043997552980768906;
          A[3][2] = 0.0001296026318932898464299230;
          A[3][3] = 4.050082246665307700935095e-06;
          A[3][4] = -1.350027415555102425806207e-06;
          A[4][1] = 9.257330849520702348385420e-06;
          A[4][2] = 4.628665424760351174192710e-06;
          A[4][3] = 8.678747671425658451611332e-07;
          A[4][4] = 4.821526484125365806450740e-08;
          A[5][2] = 1.461068631553141138430274e-08;
          A[5][3] = 7.305343157765705692151368e-09;
          A[5][4] = 1.217557192960950879760177e-09;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][0] = 0.9128709291752769017236346;
          A[0][1] = 0.4564354645876384508618173;
          A[0][2] = 0.03042903097250922866967571;
          A[0][3] = -0.02282177322938192323698026;
          A[0][4] = -0.003803628871563653583709463;
          A[0][5] = 0.0006339381452606090028983887;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.02390852433554296538331663;
          A[1][2] = 0.01195426216777148269165831;
          A[1][3] = 0.0005433755530805219405299233;
          A[1][4] = -0.0007245007374406959568466369;
          A[1][5] = 4.528129609004349730291480e-05;
          A[2][0] = 0.02173502212322087762119693;
          A[2][1] = 0.01086751106161043881059847;
          A[2][2] = 0.002535752581042436011593555;
          A[2][3] = 0.0003622503687203479784233184;
          A[2][4] = 0.000000000000000000000000;
          A[2][5] = -1.509376536334783243430493e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.0002173502212322087924750019;
          A[3][2] = 0.0001086751106161043962375010;
          A[3][3] = 2.415002458135653325014061e-05;
          A[3][4] = 3.018753072669566656267576e-06;
          A[3][5] = -5.031255121115943721251838e-07;
          A[4][0] = 0.0001552501580230062803392871;
          A[4][1] = 7.762507901150314016964354e-05;
          A[4][2] = 1.811251843601739824353956e-05;
          A[4][3] = 2.587502633716771056643802e-06;
          A[4][4] = 2.515627560557971860625919e-07;
          A[4][5] = 1.796876828969979900447085e-08;
          A[5][1] = 3.920458535934501937863562e-07;
          A[5][2] = 1.960229267967250968931781e-07;
          A[5][3] = 4.356065039927224301876792e-08;
          A[5][4] = 5.445081299909030377345991e-09;
          A[5][5] = 4.537567749924191981121659e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = -0.9128709291752769017236346;
          A[0][1] = -0.4564354645876384508618173;
          A[0][2] = -0.04564354645876384647396051;
          A[0][3] = 0.01521451548625461433483785;
          A[0][4] = 0.003803628871563653583709463;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = -0.01956151991089878985907724;
          A[1][2] = -0.009780759955449394929538620;
          A[1][3] = -0.001086751106161043881059847;
          A[1][4] = 0.0002716877765402609702649617;
          A[2][0] = -0.01086751106161043881059847;
          A[2][1] = -0.005433755530805219405299233;
          A[2][2] = -0.002173502212322087762119693;
          A[2][3] = -0.0006339381452606088944781715;
          A[2][4] = -9.056259218008699460582961e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 3.622503687203479648707913e-05;
          A[3][2] = 1.811251843601739824353956e-05;
          A[3][3] = -3.018753072669566232751103e-06;
          A[3][4] = -3.018753072669566232751103e-06;
          A[4][0] = 0.0001035001053486708490420157;
          A[4][1] = 5.175005267433542452100784e-05;
          A[4][2] = 1.164376185172547102544653e-05;
          A[4][3] = 1.509376536334783116375551e-06;
          A[4][4] = 1.078126097381987940268251e-07;
          A[5][1] = 2.613639023956334448777177e-07;
          A[5][2] = 1.306819511978167224388589e-07;
          A[5][3] = 2.722540649954515271391057e-08;
          A[5][4] = 2.722540649954515188672995e-09;
        } // |m|=1
        else if(abs(m)==2){
          A[0][0] = 0.9128709291752769017236346;
          A[0][1] = 0.4564354645876384508618173;
          A[0][2] = 0.09128709291752769294792103;
          A[0][3] = 0.007607257743127307167418927;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.006520506636966263286359080;
          A[1][2] = 0.003260253318483131643179540;
          A[1][3] = 0.0005433755530805219405299233;
          A[2][0] = -0.02173502212322087762119693;
          A[2][1] = -0.01086751106161043881059847;
          A[2][2] = -0.002173502212322087762119693;
          A[2][3] = -0.0001811251843601739892116592;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = -7.245007374406959297415826e-05;
          A[3][2] = -3.622503687203479648707913e-05;
          A[3][3] = -6.037506145339132465502206e-06;
          A[4][0] = 2.587502633716771226050392e-05;
          A[4][1] = 1.293751316858385613025196e-05;
          A[4][2] = 2.587502633716771056643802e-06;
          A[4][3] = 2.156252194763975880536502e-07;
          A[5][1] = 6.534097559890836121942944e-08;
          A[5][2] = 3.267048779945418060971472e-08;
          A[5][3] = 5.445081299909030377345991e-09;
        } // |m|=2
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][1] = 0.3240370349203930411263741;
          A[0][2] = 0.1620185174601965205631871;
          A[0][3] = 0.009001028747788695008602566;
          A[0][4] = -0.009001028747788695008602566;
          A[0][5] = 0.0007500857289823913229970254;
          A[1][0] = 0.2314550249431378825537564;
          A[1][1] = 0.1157275124715689412768782;
          A[1][2] = 0.01543033499620919193895396;
          A[1][3] = -0.001928791874526148992369246;
          A[1][4] = -0.0005357755207017080413892107;
          A[1][5] = 5.357755207017080684942650e-05;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.000000000000000000000000;
          A[2][2] = -0.000000000000000000000000;
          A[2][3] = 0.0002143102082806832273977060;
          A[2][4] = 0.0001071551041403416136988530;
          A[2][5] = -1.785918402339026894980883e-05;
          A[3][0] = 0.001714481666245465819181648;
          A[3][1] = 0.0008572408331227329095908241;
          A[3][2] = 0.0001857355138432587916570010;
          A[3][3] = 2.143102082806832341739696e-05;
          A[3][4] = -0.000000000000000000000000;
          A[3][5] = -5.953061341130089720522357e-07;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 1.428734721871221448222071e-05;
          A[4][2] = 7.143673609356107241110355e-06;
          A[4][3] = 1.530787202004880153346253e-06;
          A[4][4] = 1.700874668894311310906703e-07;
          A[4][5] = 2.126093336117889138633378e-08;
          A[5][0] = 7.731248494974142659191298e-06;
          A[5][1] = 3.865624247487071329595649e-06;
          A[5][2] = 8.762081627637361892175041e-07;
          A[5][3] = 1.159687274246121372408915e-07;
          A[5][4] = 9.664060618717678654861367e-09;
          A[5][5] = 5.368922565954265804481230e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = -0.3055050463303893293343094;
          A[0][2] = -0.1527525231651946646671547;
          A[0][3] = -0.01272937693043288930050405;
          A[0][4] = 0.006364688465216444650252026;
          A[1][0] = -0.2182178902359923899911820;
          A[1][1] = -0.1091089451179961949955910;
          A[1][2] = -0.01636634176769942786155987;
          A[1][3] = 0.0009092412093166349655246061;
          A[1][4] = 0.0004546206046583174827623031;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.001818482418633269931049212;
          A[2][2] = 0.0009092412093166349655246061;
          A[2][3] = -0.0001515402015527724942541010;
          A[2][4] = -0.0001515402015527724942541010;
          A[3][0] = -0.0006061608062110899770164041;
          A[3][1] = -0.0003030804031055449885082020;
          A[3][2] = -0.0001212321612422179899822700;
          A[3][3] = -3.535938036231358379962719e-05;
          A[3][4] = -5.051340051759082915927915e-06;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 1.298916013309478488296691e-05;
          A[4][2] = 6.494580066547392441483454e-06;
          A[4][3] = 1.443240014788309525555540e-06;
          A[4][4] = 1.804050018485386906944424e-07;
          A[5][0] = 5.466818237834505213930836e-06;
          A[5][1] = 2.733409118917252606965418e-06;
          A[5][2] = 6.013500061617956709411809e-07;
          A[5][3] = 7.289090983779340744050628e-08;
          A[5][4] = 4.555681864862087965031643e-09;
        } // |m|=1
        else if(abs(m)==2){
          A[0][1] = 0.2415229457698239712293287;
          A[0][2] = 0.1207614728849119856146643;
          A[0][3] = 0.02012691214748533324874202;
          A[1][0] = 0.1725163898355885627733386;
          A[1][1] = 0.08625819491779428138666930;
          A[1][2] = 0.01725163898355885558344447;
          A[1][3] = 0.001437636581963237965287039;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.005750546327852951861148156;
          A[2][2] = -0.002875273163926475930574078;
          A[2][3] = -0.0004792121939877459703589768;
          A[3][0] = -0.001916848775950983881435907;
          A[3][1] = -0.0009584243879754919407179536;
          A[3][2] = -0.0001916848775950983881435907;
          A[3][3] = -1.597373979959153234529923e-05;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 6.845888485539228389994797e-06;
          A[4][2] = 3.422944242769614194997398e-06;
          A[4][3] = 5.704907071282689972065269e-07;
          A[5][0] = 1.728759718570512330922204e-06;
          A[5][1] = 8.643798592852561654611022e-07;
          A[5][2] = 1.728759718570512225043086e-07;
          A[5][3] = 1.440633098808760132390531e-08;
        } // |m|=2
      } // lb=3
    } // la=2
  } // nb=3
} // na=2
else if(na==3){
  if(nb==0){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.2236067977499789638518024;
          A[0][1] = 0.1118033988749894819259012;
          A[0][2] = 0.03726779962499649628826504;
          A[0][3] = 0.009316949906249124072066259;
          A[1][1] = -0.01863389981249824814413252;
          A[1][2] = -0.009316949906249124072066259;
          A[1][3] = -0.002795084971874737134883704;
          A[2][2] = 0.0003105649968749707987882014;
          A[2][3] = 0.0001552824984374853993941007;
          A[3][3] = -2.218321406249791226308193e-06;
        } // |m|=0
      } // lb=0
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.1075828707279838092336632;
          A[0][2] = 0.05379143536399190461683162;
          A[0][3] = 0.01613743060919757207893888;
          A[1][0] = -0.06454972243679028831575550;
          A[1][1] = -0.03227486121839514415787775;
          A[1][2] = -0.01506160190191773232126771;
          A[1][3] = -0.004841229182759271103264620;
          A[2][1] = 0.001434438276373117436210403;
          A[2][2] = 0.0007172191381865587181052013;
          A[2][3] = 0.0002689571768199595057369233;
          A[3][2] = -7.684490766284557790788065e-06;
          A[3][3] = -3.842245383142278895394033e-06;
        } // |m|=0
      } // lb=0
    } // la=1
    else if(la==2){
      if(lb==0){
        if(abs(m)==0){
          A[0][2] = 0.04166666666666666435370203;
          A[0][3] = 0.02083333333333333217685102;
          A[1][1] = -0.03333333333333333287074041;
          A[1][2] = -0.01666666666666666643537020;
          A[1][3] = -0.006250000000000000346944695;
          A[2][0] = 0.008333333333333333217685102;
          A[2][1] = 0.004166666666666666608842551;
          A[2][2] = 0.001388888888888888941894328;
          A[2][3] = 0.0003472222222222222354735821;
          A[3][1] = -5.952380952380952375789513e-05;
          A[3][2] = -2.976190476190476187894757e-05;
          A[3][3] = -4.960317460317460313157928e-06;
        } // |m|=0
      } // lb=0
    } // la=2
    else if(la==3){
      if(lb==0){
        if(abs(m)==0){
          A[0][3] = 0.02465033242958173179126824;
          A[1][2] = -0.01479019945774904046253972;
          A[1][3] = -0.007395099728874520231269862;
          A[2][1] = 0.004930066485916346531725996;
          A[2][2] = 0.002465033242958173265862998;
          A[2][3] = 0.0004108388738263621929071301;
          A[3][0] = -0.0007042952122737638522009807;
          A[3][1] = -0.0003521476061368819261004903;
          A[3][2] = -7.042952122737637708858177e-05;
          A[3][3] = -5.869126768948031424048481e-06;
        } // |m|=0
      } // lb=0
    } // la=3
  } // nb=0
  else if(nb==1){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.6324555320336758823529522;
          A[0][1] = 0.3162277660168379411764761;
          A[0][2] = 0.07905694150420948529411902;
          A[0][3] = 0.01317615691736824697077868;
          A[0][4] = 0.001317615691736824783814042;
          A[1][1] = -0.02635231383473649394155736;
          A[1][2] = -0.01317615691736824697077868;
          A[1][3] = -0.002635231383473649567628083;
          A[1][4] = -0.0002196026152894707973023403;
          A[2][2] = 0.000000000000000000000000;
          A[2][3] = 0.000000000000000000000000;
          A[2][4] = -1.568590109210505598212951e-05;
          A[3][3] = 3.137180218421011196425902e-06;
          A[3][4] = 1.568590109210505598212951e-06;
          A[4][4] = -2.489825570175405911022863e-08;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = -0.000000000000000000000000;
          A[0][2] = -0.000000000000000000000000;
          A[0][3] = 0.004564354645876384126979008;
          A[0][4] = 0.002282177322938192063489504;
          A[1][0] = 0.1825741858350553581402664;
          A[1][1] = 0.09128709291752767907013322;
          A[1][2] = 0.01825741858350553650791603;
          A[1][3] = 0.001521451548625461303379525;
          A[1][4] = -0.0003803628871563653258448812;
          A[2][1] = -0.004564354645876384126979008;
          A[2][2] = -0.002282177322938192063489504;
          A[2][3] = -0.0004347004424644175307398952;
          A[2][4] = -2.716877765402609567124345e-05;
          A[3][2] = 3.260253318483131209498671e-05;
          A[3][3] = 1.630126659241565604749336e-05;
          A[3][4] = 2.716877765402609482421050e-06;
          A[4][3] = -8.625008779055902992750416e-08;
          A[4][4] = -4.312504389527951496375208e-08;
        } // |m|=0
      } // lb=1
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.1825741858350553581402664;
          A[0][2] = 0.09128709291752767907013322;
          A[0][3] = 0.01977887013213099759445512;
          A[0][4] = 0.002282177322938192063489504;
          A[1][0] = -0.1825741858350553581402664;
          A[1][1] = -0.09128709291752767907013322;
          A[1][2] = -0.02434322477800738085407239;
          A[1][3] = -0.004564354645876384126979008;
          A[1][4] = -0.0003803628871563653258448812;
          A[2][1] = 0.001521451548625461303379525;
          A[2][2] = 0.0007607257743127306516897623;
          A[2][3] = 7.245007374406959297415826e-05;
          A[2][4] = -2.716877765402609567124345e-05;
          A[3][2] = 1.086751106161043792968420e-05;
          A[3][3] = 5.433755530805218964842101e-06;
          A[3][4] = 2.716877765402609482421050e-06;
          A[4][3] = -8.625008779055902992750416e-08;
          A[4][4] = -4.312504389527951496375208e-08;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][0] = -0.6324555320336758823529522;
          A[0][1] = -0.3162277660168379411764761;
          A[0][2] = -0.04743416490252568701313507;
          A[0][3] = 0.002635231383473649567628083;
          A[0][4] = 0.003952847075210473917761256;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.04743416490252568701313507;
          A[1][2] = 0.02371708245126284350656753;
          A[1][3] = 0.002635231383473649567628083;
          A[1][4] = -0.0006588078458684123919070208;
          A[2][0] = -0.02108185106778919654102467;
          A[2][1] = -0.01054092553389459827051233;
          A[2][2] = -0.003011693009684170748568866;
          A[2][3] = -0.0006274360436842021850750717;
          A[2][4] = -4.705770327631516794638852e-05;
          A[3][1] = 0.0002258769757263128061426649;
          A[3][2] = 0.0001129384878631564030713325;
          A[3][3] = 2.823462196578910076783311e-05;
          A[3][4] = 4.705770327631516794638852e-06;
          A[4][2] = -5.975581368420973657059280e-07;
          A[4][3] = -2.987790684210486828529640e-07;
          A[4][4] = -7.469476710526217071324100e-08;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 0.6324555320336758823529522;
          A[0][1] = 0.3162277660168379411764761;
          A[0][2] = 0.06851601597031488355415974;
          A[0][3] = 0.007905694150420947835522512;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = -0.01581138830084189567104502;
          A[1][2] = -0.007905694150420947835522512;
          A[1][3] = -0.001317615691736824783814042;
          A[2][0] = -0.01054092553389459827051233;
          A[2][1] = -0.005270462766947299135256166;
          A[2][2] = -0.001066641274263143779679752;
          A[2][3] = -9.411540655263033589277705e-05;
          A[3][1] = 0.0001129384878631564030713325;
          A[3][2] = 5.646924393157820153566623e-05;
          A[3][3] = 9.411540655263033589277705e-06;
          A[4][2] = -2.987790684210486828529640e-07;
          A[4][3] = -1.493895342105243414264820e-07;
        } // |m|=1
      } // lb=1
    } // la=1
    else if(la==2){
      if(lb==0){
        if(abs(m)==0){
          A[0][2] = 0.03535533905932737586130798;
          A[0][3] = 0.01767766952966368793065399;
          A[0][4] = 0.002946278254943948422123201;
          A[1][1] = -0.05892556509887896670774055;
          A[1][2] = -0.02946278254943948335387027;
          A[1][3] = -0.005892556509887896844246402;
          A[1][4] = -0.0004910463758239913675471278;
          A[2][0] = 0.02357022603955158737698561;
          A[2][1] = 0.01178511301977579368849280;
          A[2][2] = 0.002525381361380527218962744;
          A[2][3] = 0.0002805979290422807659668991;
          A[2][4] = -3.507474113028509574586239e-05;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 7.014948226057019487985657e-06;
          A[3][4] = 3.507474113028509743992828e-06;
          A[4][2] = -6.680903072435256756061691e-07;
          A[4][3] = -3.340451536217628378030845e-07;
          A[4][4] = -5.567419227029380630051409e-08;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = -0.2449489742783177992357935;
          A[0][2] = -0.1224744871391588996178967;
          A[0][3] = -0.01020620726159657438991690;
          A[0][4] = 0.005103103630798287194958451;
          A[1][0] = 0.1632993161855451902386704;
          A[1][1] = 0.08164965809277259511933522;
          A[1][2] = 0.02041241452319314877983381;
          A[1][3] = 0.003402069087198858129972301;
          A[1][4] = -0.0008505172717997145324930752;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.004374088826398532005301778;
          A[2][2] = -0.002187044413199266002650889;
          A[2][3] = -0.0004860098695998368834546299;
          A[2][4] = -6.075123369997961043182874e-05;
          A[3][0] = 0.001749635530559412845488798;
          A[3][1] = 0.0008748177652797064227443991;
          A[3][2] = 0.0002187044413199266056860998;
          A[3][3] = 3.645074021998776761434996e-05;
          A[3][4] = 6.075123369997960704369695e-06;
          A[4][1] = -6.942998137140526761289065e-06;
          A[4][2] = -3.471499068570263380644533e-06;
          A[4][3] = -7.714442374600585290321184e-07;
          A[4][4] = -9.643052968250731612901479e-08;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.2121320343559642274122723;
          A[0][2] = 0.1060660171779821137061361;
          A[0][3] = 0.01767766952966368793065399;
          A[1][0] = -0.1414213562373095034452319;
          A[1][1] = -0.07071067811865475172261597;
          A[1][2] = -0.01767766952966368793065399;
          A[1][3] = -0.002946278254943947554761463;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.002525381361380526785281875;
          A[2][2] = -0.001262690680690263392640937;
          A[2][3] = -0.0002104484467817105473701200;
          A[3][0] = 0.001010152544552210670744663;
          A[3][1] = 0.0005050762722761053353723315;
          A[3][2] = 0.0001262690680690263338430829;
          A[3][3] = 2.104484467817105676989108e-05;
          A[4][1] = -4.008541843461153630120541e-06;
          A[4][2] = -2.004270921730576815060270e-06;
          A[4][3] = -3.340451536217627848635253e-07;
        } // |m|=1
      } // lb=1
    } // la=2
    else if(la==3){
      if(lb==0){
        if(abs(m)==0){
          A[0][3] = 0.006972166887783962790914316;
          A[0][4] = 0.003486083443891981395457158;
          A[1][2] = -0.01394433377556792558182863;
          A[1][3] = -0.006972166887783962790914316;
          A[1][4] = -0.0005810139073153303048563378;
          A[2][1] = 0.008964214570007952159746978;
          A[2][2] = 0.004482107285003976079873489;
          A[2][3] = 0.0006640158940746631590842930;
          A[2][4] = -4.150099337966644744276831e-05;
          A[3][0] = -0.001992047682223989368832662;
          A[3][1] = -0.0009960238411119946844163309;
          A[3][2] = -0.0001494035761667992135044714;
          A[3][3] = 8.300198675933289827366841e-06;
          A[3][4] = 4.150099337966644913683421e-06;
          A[4][1] = -7.904951119936466018235308e-06;
          A[4][2] = -3.952475559968233009117654e-06;
          A[4][3] = -7.904951119936465594718834e-07;
          A[4][4] = -6.587459266613721770095355e-08;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][2] = -0.07245688373094719136879860;
          A[0][3] = -0.03622844186547359568439930;
          A[0][4] = 0.006038073644245600148094955;
          A[1][1] = 0.07245688373094719136879860;
          A[1][2] = 0.03622844186547359568439930;
          A[1][3] = 0.004025382429497066476276057;
          A[1][4] = -0.001006345607374266619069014;
          A[2][0] = -0.02070196678027062808791214;
          A[2][1] = -0.01035098339013531404395607;
          A[2][2] = -0.001725163898355885601712534;
          A[2][3] = 0.000000000000000000000000;
          A[2][4] = -7.188182909816189555384652e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.0003450327796711771203425068;
          A[3][2] = 0.0001725163898355885601712534;
          A[3][3] = 4.312909745889714004281335e-05;
          A[3][4] = 7.188182909816190571824189e-06;
          A[4][0] = -0.0001095342157686276542399167;
          A[4][1] = -5.476710788431382711995837e-05;
          A[4][2] = -1.232259927397061144080381e-05;
          A[4][3] = -1.597373979959153319233217e-06;
          A[4][4] = -1.140981414256538179701511e-07;
        } // |m|=0
        else if(abs(m)==1){
          A[0][2] = 0.05916079783099616185015890;
          A[0][3] = 0.02958039891549808092507945;
          A[1][1] = -0.05916079783099616185015890;
          A[1][2] = -0.02958039891549808092507945;
          A[1][3] = -0.004930066485916346531725996;
          A[2][0] = 0.01690308509457033245282354;
          A[2][1] = 0.008451542547285166226411768;
          A[2][2] = 0.0007042952122737638522009807;
          A[2][3] = -0.0003521476061368819261004903;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.0004225771273642582896365449;
          A[3][2] = 0.0002112885636821291448182725;
          A[3][3] = 3.521476061368818854429089e-05;
          A[4][0] = -6.707573450226321433876447e-05;
          A[4][1] = -3.353786725113160716938224e-05;
          A[4][2] = -6.707573450226322111502805e-06;
          A[4][3] = -5.589644541855267720391548e-07;
        } // |m|=1
      } // lb=1
    } // la=3
  } // nb=1
  else if(nb==2){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 0.9128709291752767907013322;
          A[0][1] = 0.4564354645876383953506661;
          A[0][2] = 0.1065016084037822968744180;
          A[0][3] = 0.01521451548625461260011438;
          A[0][4] = 0.001521451548625461303379525;
          A[0][5] = 0.0001267876290521217843166452;
          A[1][1] = -0.01521451548625461260011438;
          A[1][2] = -0.007607257743127306300057189;
          A[1][3] = -0.001521451548625461303379525;
          A[1][4] = -0.0001267876290521217843166452;
          A[1][5] = -9.056259218008699121769782e-06;
          A[2][2] = -0.0002535752581042435686332903;
          A[2][3] = -0.0001267876290521217843166452;
          A[2][4] = -2.716877765402609567124345e-05;
          A[2][5] = -3.018753072669566232751103e-06;
          A[3][3] = 1.811251843601739654947367e-06;
          A[3][4] = 9.056259218008698274736835e-07;
          A[3][5] = 1.006251024223188638371249e-07;
          A[4][4] = 7.187507315879919436352217e-09;
          A[4][5] = 3.593753657939959718176109e-09;
          A[5][5] = -9.075135499848382669773610e-11;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.05270462766947298788311471;
          A[0][2] = 0.02635231383473649394155736;
          A[0][3] = 0.007905694150420947835522512;
          A[0][4] = 0.001756820922315766378418722;
          A[0][5] = 0.0002196026152894707973023403;
          A[1][0] = 0.2635231383473649602322553;
          A[1][1] = 0.1317615691736824801161276;
          A[1][2] = 0.02635231383473649394155736;
          A[1][3] = 0.002196026152894707973023403;
          A[1][4] = -3.137180218421011196425902e-05;
          A[1][5] = -1.568590109210505598212951e-05;
          A[2][1] = -0.003513641844631532756837444;
          A[2][2] = -0.001756820922315766378418722;
          A[2][3] = -0.0003764616262105213435711082;
          A[2][4] = -4.182906957894682046985441e-05;
          A[2][5] = -5.228633697368352558731801e-06;
          A[3][2] = -0.000000000000000000000000;
          A[3][3] = -0.000000000000000000000000;
          A[3][4] = 3.485755798245567878385315e-07;
          A[3][5] = 1.742877899122783939192657e-07;
          A[4][3] = 9.959302280701623644091453e-08;
          A[4][4] = 4.979651140350811822045727e-08;
          A[4][5] = 6.224563925438514777557158e-09;
          A[5][4] = -3.143719154261876171169792e-10;
          A[5][5] = -1.571859577130938085584896e-10;
        } // |m|=0
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][2] = -0.006804138174397717127306340;
          A[0][3] = -0.003402069087198858563653170;
          A[0][4] = 0.000000000000000000000000;
          A[0][5] = 0.0002835057572665715289010613;
          A[1][1] = 0.02721655269759086850922536;
          A[1][2] = 0.01360827634879543425461268;
          A[1][3] = 0.003159064152398939877980366;
          A[1][4] = 0.0004455090471331838233859379;
          A[1][5] = -2.025041123332654019874137e-05;
          A[2][0] = 0.03402069087198858476916996;
          A[2][1] = 0.01701034543599429238458498;
          A[2][2] = 0.003078062507465633649422765;
          A[2][3] = 0.0001215024673999592344161846;
          A[2][4] = -4.050082246665308039748274e-05;
          A[2][5] = -6.750137077775512552547509e-06;
          A[3][1] = -0.0004374088826398532113721995;
          A[3][2] = -0.0002187044413199266056860998;
          A[3][3] = -4.185084988220818239977247e-05;
          A[3][4] = -2.700054831110205275128888e-06;
          A[3][5] = 2.250045692591837641041475e-07;
          A[4][2] = 1.735749534285131902080503e-06;
          A[4][3] = 8.678747671425659510402516e-07;
          A[4][4] = 1.607175494708455533514709e-07;
          A[4][5] = 8.035877473542277998445791e-09;
          A[5][3] = -2.435114385921902173110660e-09;
          A[5][4] = -1.217557192960951086555330e-09;
          A[5][5] = -2.029261988268251897090197e-10;
        } // |m|=0
      } // lb=2
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.1581138830084189705882380;
          A[0][2] = 0.07905694150420948529411902;
          A[0][3] = 0.01756820922315766378418722;
          A[0][4] = 0.002196026152894707973023403;
          A[0][5] = 0.0002196026152894707973023403;
          A[1][0] = -0.2635231383473649602322553;
          A[1][1] = -0.1317615691736824801161276;
          A[1][2] = -0.02898754521821014437654718;
          A[1][3] = -0.003513641844631532756837444;
          A[1][4] = -0.0002509744174736808957140721;
          A[1][5] = -1.568590109210505598212951e-05;
          A[2][1] = -0.000000000000000000000000;
          A[2][2] = -0.000000000000000000000000;
          A[2][3] = -7.320087176315693243411342e-05;
          A[2][4] = -3.660043588157846621705671e-05;
          A[2][5] = -5.228633697368352558731801e-06;
          A[3][2] = 1.882308131052606717855541e-05;
          A[3][3] = 9.411540655263033589277705e-06;
          A[3][4] = 1.917165689035062491930601e-06;
          A[3][5] = 1.742877899122783939192657e-07;
          A[4][3] = -0.000000000000000000000000;
          A[4][4] = -0.000000000000000000000000;
          A[4][5] = 6.224563925438514777557158e-09;
          A[5][4] = -3.143719154261876171169792e-10;
          A[5][5] = -1.571859577130938085584896e-10;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][0] = -0.9128709291752767907013322;
          A[0][1] = -0.4564354645876383953506661;
          A[0][2] = -0.07302967433402214603166414;
          A[0][3] = 0.001521451548625461303379525;
          A[0][4] = 0.002282177322938192063489504;
          A[0][5] = 0.0003803628871563653258448812;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.02738612787525830302715057;
          A[1][2] = 0.01369306393762915151357529;
          A[1][3] = 0.002064827101705983027069014;
          A[1][4] = -0.0001086751106161043826849738;
          A[1][5] = -2.716877765402609567124345e-05;
          A[2][0] = -0.03042903097250922520022876;
          A[2][1] = -0.01521451548625461260011438;
          A[2][2] = -0.003477603539715340245919162;
          A[2][3] = -0.0004709254793364523340032379;
          A[2][4] = -5.433755530805219134248690e-05;
          A[2][5] = -9.056259218008699121769782e-06;
          A[3][1] = 0.0001304101327393252483799468;
          A[3][2] = 6.520506636966262418997342e-05;
          A[3][3] = 1.328251351974609023825873e-05;
          A[3][4] = 1.207501229067826366045499e-06;
          A[3][5] = 3.018753072669565915113748e-07;
          A[4][2] = 5.175005267433542325045842e-07;
          A[4][3] = 2.587502633716771162522921e-07;
          A[4][4] = 6.468756584291927906307302e-08;
          A[4][5] = 1.078126097381987874093802e-08;
          A[5][3] = -2.178032519963611737348090e-09;
          A[5][4] = -1.089016259981805868674045e-09;
          A[5][5] = -2.722540649954514671685112e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 0.9128709291752767907013322;
          A[0][1] = 0.4564354645876383953506661;
          A[0][2] = 0.1004158022092804525282617;
          A[0][3] = 0.01217161238900369042703620;
          A[0][4] = 0.0007607257743127306516897623;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = -0.009128709291752768253958017;
          A[1][2] = -0.004564354645876384126979008;
          A[1][3] = -0.0008694008849288350614797904;
          A[1][4] = -5.433755530805219134248690e-05;
          A[2][0] = -0.01521451548625461260011438;
          A[2][1] = -0.007607257743127306300057189;
          A[2][2] = -0.001738801769857670122959581;
          A[2][3] = -0.0002354627396682261670016190;
          A[2][4] = -1.811251843601739824353956e-05;
          A[3][1] = 6.520506636966262418997342e-05;
          A[3][2] = 3.260253318483131209498671e-05;
          A[3][3] = 6.641256759873045119129363e-06;
          A[3][4] = 6.037506145339131830227495e-07;
          A[4][2] = 2.587502633716771162522921e-07;
          A[4][3] = 1.293751316858385581261460e-07;
          A[4][4] = 2.156252194763975748187604e-08;
          A[5][3] = -1.089016259981805868674045e-09;
          A[5][4] = -5.445081299909029343370225e-10;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][1] = -0.1414213562373095034452319;
          A[0][2] = -0.07071067811865475172261597;
          A[0][3] = -0.01374929852307175785763871;
          A[0][4] = -0.0009820927516479827350942555;
          A[0][5] = 0.0004910463758239913675471278;
          A[1][0] = -0.2357022603955158668309622;
          A[1][1] = -0.1178511301977579334154811;
          A[1][2] = -0.01431049438115632004009381;
          A[1][3] = 0.002665680325901667330895650;
          A[1][4] = 0.0007014948226057019691273564;
          A[1][5] = -3.507474113028509574586239e-05;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.004208968935634212031604573;
          A[2][2] = 0.002104484467817106015802286;
          A[2][3] = 0.0002572147682887573959080452;
          A[2][4] = -4.676632150704679432781652e-05;
          A[2][5] = -1.169158037676169858195413e-05;
          A[3][0] = -0.002525381361380527218962744;
          A[3][1] = -0.001262690680690263609481372;
          A[3][2] = -0.0003086577219465088642476325;
          A[3][3] = -4.910463758239913810996549e-05;
          A[3][4] = -3.897193458920566758673341e-06;
          A[3][5] = 3.897193458920566441035986e-07;
          A[4][1] = 1.603416737384461452048216e-05;
          A[4][2] = 8.017083686922307260241082e-06;
          A[4][3] = 1.837248344919695660856524e-06;
          A[4][4] = 2.505338652163221018825338e-07;
          A[4][5] = 1.391854806757345157512852e-08;
          A[5][2] = -2.530645103195173043739026e-08;
          A[5][3] = -1.265322551597586521869513e-08;
          A[5][4] = -2.811827892439081113755551e-09;
          A[5][5] = -3.514784865548851392194438e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.1224744871391588996178967;
          A[0][2] = 0.06123724356957944980894837;
          A[0][3] = 0.01360827634879543251988920;
          A[0][4] = 0.001701034543599429064986150;
          A[1][0] = 0.2041241452319314808594442;
          A[1][1] = 0.1020620726159657404297221;
          A[1][2] = 0.01895438491439363840052046;
          A[1][3] = 0.0009720197391996737669092599;
          A[1][4] = -0.0001215024673999592208636575;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.004374088826398532005301778;
          A[2][2] = -0.002187044413199266002650889;
          A[2][3] = -0.0004455090471331838233859379;
          A[2][4] = -4.050082246665307362121916e-05;
          A[3][0] = -0.001458029608799510596153781;
          A[3][1] = -0.0007290148043997552980768906;
          A[3][2] = -0.0001296026318932898464299230;
          A[3][3] = -4.050082246665307700935095e-06;
          A[3][4] = 1.350027415555102425806207e-06;
          A[4][1] = 9.257330849520702348385420e-06;
          A[4][2] = 4.628665424760351174192710e-06;
          A[4][3] = 8.678747671425658451611332e-07;
          A[4][4] = 4.821526484125365806450740e-08;
          A[5][2] = -1.461068631553141138430274e-08;
          A[5][3] = -7.305343157765705692151368e-09;
          A[5][4] = -1.217557192960950879760177e-09;
        } // |m|=1
      } // lb=2
    } // la=1
    else if(la==2){
      if(lb==0){
        if(abs(m)==0){
          A[0][2] = 0.01360827634879543425461268;
          A[0][3] = 0.006804138174397717127306340;
          A[0][4] = 0.001701034543599429281826585;
          A[0][5] = 0.0002835057572665715289010613;
          A[1][1] = -0.05443310539518173701845072;
          A[1][2] = -0.02721655269759086850922536;
          A[1][3] = -0.005346108565598205880631255;
          A[1][4] = -0.0004050082246665307633172459;
          A[1][5] = -2.025041123332654019874137e-05;
          A[2][0] = 0.03402069087198858476916996;
          A[2][1] = 0.01701034543599429238458498;
          A[2][2] = 0.003321067442265552335095569;
          A[2][3] = 0.0002430049347999184688323693;
          A[2][4] = -2.025041123332654019874137e-05;
          A[2][5] = -6.750137077775512552547509e-06;
          A[3][1] = 0.0001458029608799510704573998;
          A[3][2] = 7.290148043997553522869992e-05;
          A[3][3] = 1.890038381777143480831985e-05;
          A[3][4] = 3.375068538887756276273755e-06;
          A[3][5] = 2.250045692591837641041475e-07;
          A[4][2] = -5.785831780950440026532072e-07;
          A[4][3] = -2.892915890475220013266036e-07;
          A[4][4] = -3.214350989416911199378317e-08;
          A[4][5] = 8.035877473542277998445791e-09;
          A[5][3] = -2.435114385921902173110660e-09;
          A[5][4] = -1.217557192960951086555330e-09;
          A[5][5] = -2.029261988268251897090197e-10;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = -0.2357022603955158668309622;
          A[0][2] = -0.1178511301977579334154811;
          A[0][3] = -0.01767766952966368793065399;
          A[0][4] = 0.0009820927516479827350942555;
          A[0][5] = 0.0004910463758239913675471278;
          A[1][0] = 0.2357022603955158668309622;
          A[1][1] = 0.1178511301977579334154811;
          A[1][2] = 0.02104484467817105755593765;
          A[1][3] = 0.0007014948226057019691273564;
          A[1][4] = -0.0002805979290422807659668991;
          A[1][5] = -3.507474113028509574586239e-05;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.001964185503295965470188511;
          A[2][2] = -0.0009820927516479827350942555;
          A[2][3] = -0.0002104484467817105744751743;
          A[2][4] = -2.338316075352339716390826e-05;
          A[2][5] = -1.169158037676169858195413e-05;
          A[3][0] = 0.002525381361380527218962744;
          A[3][1] = 0.001262690680690263609481372;
          A[3][2] = 0.0002805979290422807659668991;
          A[3][3] = 3.507474113028509574586239e-05;
          A[3][4] = 3.117754767136453152828789e-06;
          A[3][5] = 3.897193458920566441035986e-07;
          A[4][1] = -0.000000000000000000000000;
          A[4][2] = -0.000000000000000000000000;
          A[4][3] = 5.567419227029380630051409e-08;
          A[4][4] = 2.783709613514690315025704e-08;
          A[4][5] = 1.391854806757345157512852e-08;
          A[5][2] = -2.530645103195173043739026e-08;
          A[5][3] = -1.265322551597586521869513e-08;
          A[5][4] = -2.811827892439081113755551e-09;
          A[5][5] = -3.514784865548851392194438e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.2041241452319314808594442;
          A[0][2] = 0.1020620726159657404297221;
          A[0][3] = 0.02041241452319314877983381;
          A[0][4] = 0.001701034543599429064986150;
          A[1][0] = -0.2041241452319314808594442;
          A[1][1] = -0.1020620726159657404297221;
          A[1][2] = -0.02187044413199265915914715;
          A[1][3] = -0.002430049347999184254642824;
          A[1][4] = -0.0001215024673999592208636575;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.003402069087198858129972301;
          A[2][2] = -0.001701034543599429064986150;
          A[2][3] = -0.0003645074021998776490384453;
          A[2][4] = -4.050082246665307362121916e-05;
          A[3][0] = 0.001458029608799510596153781;
          A[3][1] = 0.0007290148043997552980768906;
          A[3][2] = 0.0001620032898666122944848766;
          A[3][3] = 2.025041123332653681060958e-05;
          A[3][4] = 1.350027415555102425806207e-06;
          A[4][1] = -0.000000000000000000000000;
          A[4][2] = -0.000000000000000000000000;
          A[4][3] = 9.643052968250731612901479e-08;
          A[4][4] = 4.821526484125365806450740e-08;
          A[5][2] = -1.461068631553141138430274e-08;
          A[5][3] = -7.305343157765705692151368e-09;
          A[5][4] = -1.217557192960950879760177e-09;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][0] = 0.9128709291752769017236346;
          A[0][1] = 0.4564354645876384508618173;
          A[0][2] = 0.03042903097250922866967571;
          A[0][3] = -0.02282177322938192323698026;
          A[0][4] = -0.003803628871563653583709463;
          A[0][5] = 0.0006339381452606090028983887;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = -0.02390852433554296538331663;
          A[1][2] = -0.01195426216777148269165831;
          A[1][3] = -0.0005433755530805219405299233;
          A[1][4] = 0.0007245007374406959568466369;
          A[1][5] = -4.528129609004349730291480e-05;
          A[2][0] = 0.02173502212322087762119693;
          A[2][1] = 0.01086751106161043881059847;
          A[2][2] = 0.002535752581042436011593555;
          A[2][3] = 0.0003622503687203479784233184;
          A[2][4] = 0.000000000000000000000000;
          A[2][5] = -1.509376536334783243430493e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = -0.0002173502212322087924750019;
          A[3][2] = -0.0001086751106161043962375010;
          A[3][3] = -2.415002458135653325014061e-05;
          A[3][4] = -3.018753072669566656267576e-06;
          A[3][5] = 5.031255121115943721251838e-07;
          A[4][0] = 0.0001552501580230062803392871;
          A[4][1] = 7.762507901150314016964354e-05;
          A[4][2] = 1.811251843601739824353956e-05;
          A[4][3] = 2.587502633716771056643802e-06;
          A[4][4] = 2.515627560557971860625919e-07;
          A[4][5] = 1.796876828969979900447085e-08;
          A[5][1] = -3.920458535934501937863562e-07;
          A[5][2] = -1.960229267967250968931781e-07;
          A[5][3] = -4.356065039927224301876792e-08;
          A[5][4] = -5.445081299909030377345991e-09;
          A[5][5] = -4.537567749924191981121659e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = -0.9128709291752769017236346;
          A[0][1] = -0.4564354645876384508618173;
          A[0][2] = -0.04564354645876384647396051;
          A[0][3] = 0.01521451548625461433483785;
          A[0][4] = 0.003803628871563653583709463;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.01956151991089878985907724;
          A[1][2] = 0.009780759955449394929538620;
          A[1][3] = 0.001086751106161043881059847;
          A[1][4] = -0.0002716877765402609702649617;
          A[2][0] = -0.01086751106161043881059847;
          A[2][1] = -0.005433755530805219405299233;
          A[2][2] = -0.002173502212322087762119693;
          A[2][3] = -0.0006339381452606088944781715;
          A[2][4] = -9.056259218008699460582961e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = -3.622503687203479648707913e-05;
          A[3][2] = -1.811251843601739824353956e-05;
          A[3][3] = 3.018753072669566232751103e-06;
          A[3][4] = 3.018753072669566232751103e-06;
          A[4][0] = 0.0001035001053486708490420157;
          A[4][1] = 5.175005267433542452100784e-05;
          A[4][2] = 1.164376185172547102544653e-05;
          A[4][3] = 1.509376536334783116375551e-06;
          A[4][4] = 1.078126097381987940268251e-07;
          A[5][1] = -2.613639023956334448777177e-07;
          A[5][2] = -1.306819511978167224388589e-07;
          A[5][3] = -2.722540649954515271391057e-08;
          A[5][4] = -2.722540649954515188672995e-09;
        } // |m|=1
        else if(abs(m)==2){
          A[0][0] = 0.9128709291752769017236346;
          A[0][1] = 0.4564354645876384508618173;
          A[0][2] = 0.09128709291752769294792103;
          A[0][3] = 0.007607257743127307167418927;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = -0.006520506636966263286359080;
          A[1][2] = -0.003260253318483131643179540;
          A[1][3] = -0.0005433755530805219405299233;
          A[2][0] = -0.02173502212322087762119693;
          A[2][1] = -0.01086751106161043881059847;
          A[2][2] = -0.002173502212322087762119693;
          A[2][3] = -0.0001811251843601739892116592;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 7.245007374406959297415826e-05;
          A[3][2] = 3.622503687203479648707913e-05;
          A[3][3] = 6.037506145339132465502206e-06;
          A[4][0] = 2.587502633716771226050392e-05;
          A[4][1] = 1.293751316858385613025196e-05;
          A[4][2] = 2.587502633716771056643802e-06;
          A[4][3] = 2.156252194763975880536502e-07;
          A[5][1] = -6.534097559890836121942944e-08;
          A[5][2] = -3.267048779945418060971472e-08;
          A[5][3] = -5.445081299909030377345991e-09;
        } // |m|=2
      } // lb=2
    } // la=2
    else if(la==3){
      if(lb==0){
        if(abs(m)==0){
          A[0][3] = -0.000000000000000000000000;
          A[0][4] = -0.000000000000000000000000;
          A[0][5] = 0.0003354485357914222063563381;
          A[1][2] = -0.006900655593423542406850135;
          A[1][3] = -0.003450327796711771203425068;
          A[1][4] = -0.0006229758521840697343616156;
          A[1][5] = -2.396060969938729851794884e-05;
          A[2][1] = 0.008625819491779427791722235;
          A[2][2] = 0.004312909745889713895861117;
          A[2][3] = 0.0007667395103803935525743629;
          A[2][4] = 2.396060969938729851794884e-05;
          A[2][5] = -7.986869899795766172649614e-06;
          A[3][0] = -0.002875273163926475930574078;
          A[3][1] = -0.001437636581963237965287039;
          A[3][2] = -0.0002268271051541997484612273;
          A[3][3] = 6.389495919836613276932870e-06;
          A[3][4] = 5.590808929857035982041551e-06;
          A[3][5] = 2.662289966598588688923498e-07;
          A[4][1] = -1.825570262810460791060886e-05;
          A[4][2] = -9.127851314052303955304431e-06;
          A[4][3] = -1.711472121384807097498699e-06;
          A[4][4] = -9.508178452137817502434769e-08;
          A[4][5] = 9.508178452137816509818034e-09;
          A[5][2] = -2.881266197617520264781062e-08;
          A[5][3] = -1.440633098808760132390531e-08;
          A[5][4] = -2.881266197617520430217184e-09;
          A[5][5] = -2.401055164681266680522398e-10;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][2] = -0.04183300132670377674548590;
          A[0][3] = -0.02091650066335188837274295;
          A[0][4] = -0.002324055629261321219425351;
          A[0][5] = 0.0005810139073153303048563378;
          A[1][1] = 0.07171371656006361727797582;
          A[1][2] = 0.03585685828003180863898791;
          A[1][3] = 0.004814115232041307930466179;
          A[1][4] = -0.0005810139073153303048563378;
          A[1][5] = -4.150099337966644744276831e-05;
          A[2][0] = -0.02988071523335984053248993;
          A[2][1] = -0.01494035761667992026624496;
          A[2][2] = -0.001992047682223989368832662;
          A[2][3] = 0.0002490059602779986711040827;
          A[2][4] = 5.533465783955525873951536e-05;
          A[2][5] = -1.383366445988881468487884e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.0001438701103828436700122345;
          A[3][2] = 7.193505519142183500611726e-05;
          A[3][3] = 2.490059602779986778803463e-05;
          A[3][4] = 6.455710081281447361163228e-06;
          A[3][5] = 4.611221486629605106717851e-07;
          A[4][0] = -0.0001580990223987293068121790;
          A[4][1] = -7.904951119936465340608950e-05;
          A[4][2] = -1.699564490786340176979932e-05;
          A[4][3] = -1.910363187317979485381220e-06;
          A[4][4] = -6.587459266613721770095355e-08;
          A[4][5] = 1.646864816653430442523839e-08;
          A[5][1] = -3.992399555523467595071114e-07;
          A[5][2] = -1.996199777761733797535557e-07;
          A[5][3] = -4.491449499963900845931657e-08;
          A[5][4] = -5.822249351805057019769457e-09;
          A[5][5] = -4.158749537003612009267360e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][2] = 0.03415650255319865880032637;
          A[0][3] = 0.01707825127659932940016319;
          A[0][4] = 0.002846375212766554755466908;
          A[1][1] = -0.05855400437691198750611576;
          A[1][2] = -0.02927700218845599375305788;
          A[1][3] = -0.005286125395137887712782021;
          A[1][4] = -0.0002033125151976110617062232;
          A[2][0] = 0.02439750182371332870578939;
          A[2][1] = 0.01219875091185666435289470;
          A[2][2] = 0.001626500121580888493649786;
          A[2][3] = -0.0002033125151976110617062232;
          A[2][4] = -6.777083839920368271789869e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.0005150583718339480591291712;
          A[3][2] = 0.0002575291859169740295645856;
          A[3][3] = 4.743958687944258332353994e-05;
          A[3][4] = 2.259027946640122954904311e-06;
          A[4][0] = -9.681548342743384019844222e-05;
          A[4][1] = -4.840774171371692009922111e-05;
          A[4][2] = -8.713393508469046227723522e-06;
          A[4][3] = -3.227182780914461212893132e-07;
          A[4][4] = 8.067956952286153032232829e-08;
          A[5][1] = -2.444835440086713088197426e-07;
          A[5][2] = -1.222417720043356544098713e-07;
          A[5][3] = -2.444835440086713220546324e-08;
          A[5][4] = -2.037362866738927545925168e-09;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][1] = 0.3240370349203930411263741;
          A[0][2] = 0.1620185174601965205631871;
          A[0][3] = 0.009001028747788695008602566;
          A[0][4] = -0.009001028747788695008602566;
          A[0][5] = 0.0007500857289823913229970254;
          A[1][0] = -0.2314550249431378825537564;
          A[1][1] = -0.1157275124715689412768782;
          A[1][2] = -0.01543033499620919193895396;
          A[1][3] = 0.001928791874526148992369246;
          A[1][4] = 0.0005357755207017080413892107;
          A[1][5] = -5.357755207017080684942650e-05;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.000000000000000000000000;
          A[2][2] = -0.000000000000000000000000;
          A[2][3] = 0.0002143102082806832273977060;
          A[2][4] = 0.0001071551041403416136988530;
          A[2][5] = -1.785918402339026894980883e-05;
          A[3][0] = -0.001714481666245465819181648;
          A[3][1] = -0.0008572408331227329095908241;
          A[3][2] = -0.0001857355138432587916570010;
          A[3][3] = -2.143102082806832341739696e-05;
          A[3][4] = -0.000000000000000000000000;
          A[3][5] = 5.953061341130089720522357e-07;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 1.428734721871221448222071e-05;
          A[4][2] = 7.143673609356107241110355e-06;
          A[4][3] = 1.530787202004880153346253e-06;
          A[4][4] = 1.700874668894311310906703e-07;
          A[4][5] = 2.126093336117889138633378e-08;
          A[5][0] = -7.731248494974142659191298e-06;
          A[5][1] = -3.865624247487071329595649e-06;
          A[5][2] = -8.762081627637361892175041e-07;
          A[5][3] = -1.159687274246121372408915e-07;
          A[5][4] = -9.664060618717678654861367e-09;
          A[5][5] = -5.368922565954265804481230e-10;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = -0.3055050463303893293343094;
          A[0][2] = -0.1527525231651946646671547;
          A[0][3] = -0.01272937693043288930050405;
          A[0][4] = 0.006364688465216444650252026;
          A[1][0] = 0.2182178902359923899911820;
          A[1][1] = 0.1091089451179961949955910;
          A[1][2] = 0.01636634176769942786155987;
          A[1][3] = -0.0009092412093166349655246061;
          A[1][4] = -0.0004546206046583174827623031;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.001818482418633269931049212;
          A[2][2] = 0.0009092412093166349655246061;
          A[2][3] = -0.0001515402015527724942541010;
          A[2][4] = -0.0001515402015527724942541010;
          A[3][0] = 0.0006061608062110899770164041;
          A[3][1] = 0.0003030804031055449885082020;
          A[3][2] = 0.0001212321612422179899822700;
          A[3][3] = 3.535938036231358379962719e-05;
          A[3][4] = 5.051340051759082915927915e-06;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 1.298916013309478488296691e-05;
          A[4][2] = 6.494580066547392441483454e-06;
          A[4][3] = 1.443240014788309525555540e-06;
          A[4][4] = 1.804050018485386906944424e-07;
          A[5][0] = -5.466818237834505213930836e-06;
          A[5][1] = -2.733409118917252606965418e-06;
          A[5][2] = -6.013500061617956709411809e-07;
          A[5][3] = -7.289090983779340744050628e-08;
          A[5][4] = -4.555681864862087965031643e-09;
        } // |m|=1
        else if(abs(m)==2){
          A[0][1] = 0.2415229457698239712293287;
          A[0][2] = 0.1207614728849119856146643;
          A[0][3] = 0.02012691214748533324874202;
          A[1][0] = -0.1725163898355885627733386;
          A[1][1] = -0.08625819491779428138666930;
          A[1][2] = -0.01725163898355885558344447;
          A[1][3] = -0.001437636581963237965287039;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.005750546327852951861148156;
          A[2][2] = -0.002875273163926475930574078;
          A[2][3] = -0.0004792121939877459703589768;
          A[3][0] = 0.001916848775950983881435907;
          A[3][1] = 0.0009584243879754919407179536;
          A[3][2] = 0.0001916848775950983881435907;
          A[3][3] = 1.597373979959153234529923e-05;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 6.845888485539228389994797e-06;
          A[4][2] = 3.422944242769614194997398e-06;
          A[4][3] = 5.704907071282689972065269e-07;
          A[5][0] = -1.728759718570512330922204e-06;
          A[5][1] = -8.643798592852561654611022e-07;
          A[5][2] = -1.728759718570512225043086e-07;
          A[5][3] = -1.440633098808760132390531e-08;
        } // |m|=2
      } // lb=2
    } // la=3
  } // nb=2
  else if(nb==3){
    if(la==0){
      if(lb==0){
        if(abs(m)==0){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.1166666666666666685170384;
          A[0][3] = 0.01666666666666666643537020;
          A[0][4] = 0.001666666666666666773641281;
          A[0][5] = 0.0001388888888888888887684220;
          A[0][6] = 9.920634920634920626315856e-06;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[1][5] = 0.000000000000000000000000;
          A[1][6] = 0.000000000000000000000000;
          A[2][2] = -0.0002777777777777777775368440;
          A[2][3] = -0.0001388888888888888887684220;
          A[2][4] = -2.976190476190476187894757e-05;
          A[2][5] = -3.306878306878307016610776e-06;
          A[2][6] = -2.755731922398589356974178e-07;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[3][5] = 0.000000000000000000000000;
          A[3][6] = 0.000000000000000000000000;
          A[4][4] = 7.873519778281683829801616e-09;
          A[4][5] = 3.936759889140841914900808e-09;
          A[4][6] = 5.368308939737511890133059e-10;
          A[5][5] = 0.000000000000000000000000;
          A[5][6] = 0.000000000000000000000000;
          A[6][6] = -1.737991757231776501695963e-13;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = 0.09622504486493763109500321;
          A[0][2] = 0.04811252243246881554750161;
          A[0][3] = 0.01154700538379251593956720;
          A[0][4] = 0.001764125822523856601878323;
          A[0][5] = 0.0001947411622266594717796384;
          A[0][6] = 1.718304372588171869610900e-05;
          A[1][0] = 0.2886751345948128655294340;
          A[1][1] = 0.1443375672974064327647170;
          A[1][2] = 0.03079201435678004250551254;
          A[1][3] = 0.003367876570272816932200000;
          A[1][4] = 0.0001718304372588171937373536;
          A[1][5] = 5.727681241960572898703000e-06;
          A[1][6] = -0.000000000000000000000000;
          A[2][1] = -0.001603750747748960547162111;
          A[2][2] = -0.0008018753738744802735810557;
          A[2][3] = -0.0002061965247105806379058351;
          A[2][4] = -3.627531453241696507991745e-05;
          A[2][5] = -4.773067701633810748919166e-06;
          A[2][6] = -4.773067701633810748919166e-07;
          A[3][2] = -1.374643498070537563451356e-05;
          A[3][3] = -6.873217490352687817256779e-06;
          A[3][4] = -1.272818053769016199711778e-06;
          A[3][5] = -6.364090268845081263256685e-08;
          A[3][6] = -0.000000000000000000000000;
          A[4][3] = 5.454934516152927080681940e-08;
          A[4][4] = 2.727467258076463540340970e-08;
          A[4][5] = 6.405415530331087865485054e-09;
          A[4][6] = 9.298183834351579593177819e-10;
          A[5][4] = 1.147923930166861751569615e-10;
          A[5][5] = 5.739619650834308757848077e-11;
          A[5][6] = -0.000000000000000000000000;
          A[6][5] = -6.020580053322700922294495e-13;
          A[6][6] = -3.010290026661350461147248e-13;
        } // |m|=0
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][2] = 0.000000000000000000000000;
          A[0][3] = 0.000000000000000000000000;
          A[0][4] = 0.0002661985687499749471569832;
          A[0][5] = 0.0001330992843749874735784916;
          A[0][6] = 2.218321406249791226308193e-05;
          A[1][1] = 0.03726779962499649628826504;
          A[1][2] = 0.01863389981249824814413252;
          A[1][3] = 0.003992978531249624478405291;
          A[1][4] = 0.0004436642812499582452616387;
          A[1][5] = 2.218321406249791226308193e-05;
          A[1][6] = 0.000000000000000000000000;
          A[2][0] = 0.03726779962499649628826504;
          A[2][1] = 0.01863389981249824814413252;
          A[2][2] = 0.003726779962499649368617982;
          A[2][3] = 0.0003105649968749707987882014;
          A[2][4] = -7.394404687499304370038294e-06;
          A[2][5] = -3.697202343749652185019147e-06;
          A[2][6] = -6.162003906249420308365245e-07;
          A[3][1] = -0.0002661985687499749471569832;
          A[3][2] = -0.0001330992843749874735784916;
          A[3][3] = -2.957761874999721748015317e-05;
          A[3][4] = -3.697202343749652185019147e-06;
          A[3][5] = -2.464801562499767911587861e-07;
          A[3][6] = 0.000000000000000000000000;
          A[4][2] = 0.000000000000000000000000;
          A[4][3] = 0.000000000000000000000000;
          A[4][4] = 1.440468445616747530214085e-08;
          A[4][5] = 7.202342228083737651070426e-09;
          A[4][6] = 1.200390371347289677443455e-09;
          A[5][3] = 2.667534158549532708449969e-09;
          A[5][4] = 1.333767079274766354224985e-09;
          A[5][5] = 2.222945132124610590374974e-10;
          A[5][6] = 0.000000000000000000000000;
          A[6][4] = -4.663521256205476864704731e-12;
          A[6][5] = -2.331760628102738432352365e-12;
          A[6][6] = -3.886267713504564053920609e-13;
        } // |m|=0
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][3] = -0.001259881576697424211686016;
          A[0][4] = -0.0006299407883487121058430080;
          A[0][5] = -5.249506569572601107900520e-05;
          A[0][6] = 2.624753284786300553950260e-05;
          A[1][2] = 0.001889822365046136317529024;
          A[1][3] = 0.0009449111825230681587645121;
          A[1][4] = 0.0002624753284786300621712896;
          A[1][5] = 5.249506569572601107900520e-05;
          A[1][6] = -0.000000000000000000000000;
          A[2][1] = 0.006299407883487121058430080;
          A[2][2] = 0.003149703941743560529215040;
          A[2][3] = 0.0006474391435806207874964491;
          A[2][4] = 6.124424331168034512946213e-05;
          A[2][5] = -0.000000000000000000000000;
          A[2][6] = -7.290981346628612414574903e-07;
          A[3][0] = 0.003149703941743560529215040;
          A[3][1] = 0.001574851970871780264607520;
          A[3][2] = 0.0002764740126641569641259555;
          A[3][3] = 6.999342092763467917991907e-06;
          A[3][4] = -4.374588807977167025228468e-06;
          A[3][5] = -5.832785077302889931659922e-07;
          A[3][6] = -0.000000000000000000000000;
          A[4][1] = -2.499765033129809873907630e-05;
          A[4][2] = -1.249882516564904936953815e-05;
          A[4][3] = -2.454314759800177007150867e-06;
          A[4][4] = -1.855886160960010384674012e-07;
          A[4][5] = 6.628164860571465305044068e-09;
          A[4][6] = 1.420321041551028368278794e-09;
          A[5][2] = 6.312537962449014418674233e-08;
          A[5][3] = 3.156268981224507209337117e-08;
          A[5][4] = 6.312537962449015245854846e-09;
          A[5][5] = 5.260448302040845693553783e-10;
          A[5][6] = -0.000000000000000000000000;
          A[6][3] = -5.517952764378509550106957e-11;
          A[6][4] = -2.758976382189254775053478e-11;
          A[6][5] = -5.517952764378509711665670e-12;
          A[6][6] = -4.598293970315424423140739e-13;
        } // |m|=0
      } // lb=3
    } // la=0
    else if(la==1){
      if(lb==0){
        if(abs(m)==0){
          A[0][1] = 0.09622504486493763109500321;
          A[0][2] = 0.04811252243246881554750161;
          A[0][3] = 0.01154700538379251593956720;
          A[0][4] = 0.001764125822523856601878323;
          A[0][5] = 0.0001947411622266594717796384;
          A[0][6] = 1.718304372588171869610900e-05;
          A[1][0] = -0.2886751345948128655294340;
          A[1][1] = -0.1443375672974064327647170;
          A[1][2] = -0.03079201435678004250551254;
          A[1][3] = -0.003367876570272816932200000;
          A[1][4] = -0.0001718304372588171937373536;
          A[1][5] = -5.727681241960572898703000e-06;
          A[1][6] = -0.000000000000000000000000;
          A[2][1] = -0.001603750747748960547162111;
          A[2][2] = -0.0008018753738744802735810557;
          A[2][3] = -0.0002061965247105806379058351;
          A[2][4] = -3.627531453241696507991745e-05;
          A[2][5] = -4.773067701633810748919166e-06;
          A[2][6] = -4.773067701633810748919166e-07;
          A[3][2] = 1.374643498070537563451356e-05;
          A[3][3] = 6.873217490352687817256779e-06;
          A[3][4] = 1.272818053769016199711778e-06;
          A[3][5] = 6.364090268845081263256685e-08;
          A[3][6] = -0.000000000000000000000000;
          A[4][3] = 5.454934516152927080681940e-08;
          A[4][4] = 2.727467258076463540340970e-08;
          A[4][5] = 6.405415530331087865485054e-09;
          A[4][6] = 9.298183834351579593177819e-10;
          A[5][4] = -1.147923930166861751569615e-10;
          A[5][5] = -5.739619650834308757848077e-11;
          A[5][6] = -0.000000000000000000000000;
          A[6][5] = -6.020580053322700922294495e-13;
          A[6][6] = -3.010290026661350461147248e-13;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][0] = -1.000000000000000000000000;
          A[0][1] = -0.5000000000000000000000000;
          A[0][2] = -0.08999999999999999666933093;
          A[0][3] = -0.003333333333333333113601693;
          A[0][4] = 0.001349206349206349096758739;
          A[0][5] = 0.0002579365079365079091791579;
          A[0][6] = 2.976190476190476187894757e-05;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[1][5] = 0.000000000000000000000000;
          A[1][6] = 0.000000000000000000000000;
          A[2][0] = -0.03333333333333333287074041;
          A[2][1] = -0.01666666666666666643537020;
          A[2][2] = -0.003690476190476190201938955;
          A[2][3] = -0.0004563492063492063217054751;
          A[2][4] = -4.298941798941798486319299e-05;
          A[2][5] = -6.613756613756613186188606e-06;
          A[2][6] = -8.267195767195766482735757e-07;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[3][5] = 0.000000000000000000000000;
          A[3][6] = 0.000000000000000000000000;
          A[4][2] = 7.558578987150415947213959e-07;
          A[4][3] = 3.779289493575207973606980e-07;
          A[4][4] = 8.374561945990517638708898e-08;
          A[4][5] = 1.037873061682585452869925e-08;
          A[4][6] = 1.610492681921253360244765e-09;
          A[5][3] = 0.000000000000000000000000;
          A[5][4] = 0.000000000000000000000000;
          A[5][5] = 0.000000000000000000000000;
          A[5][6] = 0.000000000000000000000000;
          A[6][4] = -4.171180217356263604070311e-12;
          A[6][5] = -2.085590108678131802035155e-12;
          A[6][6] = -5.213975271695329505087888e-13;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.1133333333333333275971810;
          A[0][3] = 0.01499999999999999944488849;
          A[0][4] = 0.001230158730158730211873275;
          A[0][5] = 5.952380952380952375789513e-05;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[1][5] = 0.000000000000000000000000;
          A[2][0] = -0.01666666666666666643537020;
          A[2][1] = -0.008333333333333333217685102;
          A[2][2] = -0.001984126984126984016842954;
          A[2][3] = -0.0002976190476190475916844214;
          A[2][4] = -2.976190476190476187894757e-05;
          A[2][5] = -1.653439153439153296547151e-06;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[3][5] = 0.000000000000000000000000;
          A[4][2] = 3.779289493575207973606980e-07;
          A[4][3] = 1.889644746787603986803490e-07;
          A[4][4] = 3.793604984081174379710185e-08;
          A[4][5] = 3.220985363842506720489529e-09;
          A[5][3] = 0.000000000000000000000000;
          A[5][4] = 0.000000000000000000000000;
          A[5][5] = 0.000000000000000000000000;
          A[6][4] = -2.085590108678131802035155e-12;
          A[6][5] = -1.042795054339065901017578e-12;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][1] = -0.1721325931647740836716309;
          A[0][2] = -0.08606629658238704183581547;
          A[0][3] = -0.01659850005517464477389211;
          A[0][4] = -0.001127058645721735205894043;
          A[0][5] = 0.0001280748461047426321052223;
          A[0][6] = 3.842245383142278556580854e-05;
          A[1][0] = -0.2581988897471611532630220;
          A[1][1] = -0.1290994448735805766315110;
          A[1][2] = -0.02182395377624814225558936;
          A[1][3] = -0.0001536898153256911422632341;
          A[1][4] = 0.0003842245383142278963156668;
          A[1][5] = 2.561496922094852371053902e-05;
          A[1][6] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.001024598768837941056841778;
          A[2][2] = 0.0005122993844189705284208891;
          A[2][3] = 7.684490766284557113161707e-05;
          A[2][4] = -4.269161536824754234100820e-06;
          A[2][5] = -4.269161536824754234100820e-06;
          A[2][6] = -1.067290384206188558525205e-06;
          A[3][0] = -0.002766416675862440940208975;
          A[3][1] = -0.001383208337931220470104487;
          A[3][2] = -0.0003125026244955719919790815;
          A[3][3] = -4.098395075351763929211515e-05;
          A[3][4] = -3.557634614020628669589508e-06;
          A[3][5] = -2.846107691216502787440840e-07;
          A[3][6] = -0.000000000000000000000000;
          A[4][1] = 7.318562634556721181158714e-06;
          A[4][2] = 3.659281317278360590579357e-06;
          A[4][3] = 8.205661135715111588858272e-07;
          A[4][4] = 1.053429470125588635613005e-07;
          A[4][5] = 1.155076173383320953649549e-08;
          A[4][6] = 2.079137112089977766200024e-09;
          A[5][2] = 1.848121877413313459664829e-08;
          A[5][3] = 9.240609387066567298324144e-09;
          A[5][4] = 2.053468752681459353673109e-09;
          A[5][5] = 2.566835940851824192091386e-10;
          A[5][6] = -0.000000000000000000000000;
          A[6][3] = -4.846473454755192770035815e-11;
          A[6][4] = -2.423236727377596385017907e-11;
          A[6][5] = -5.384970505283547612016857e-12;
          A[6][6] = -6.731213131604434515021072e-13;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.1490711984999859851530601;
          A[0][2] = 0.07453559924999299257653007;
          A[0][3] = 0.01650431126249844726583405;
          A[0][4] = 0.002040855693749807873993429;
          A[0][5] = 0.0001330992843749874735784916;
          A[1][0] = 0.2236067977499789638518024;
          A[1][1] = 0.1118033988749894819259012;
          A[1][2] = 0.02342547404999779578349539;
          A[1][3] = 0.002395787118749774687043175;
          A[1][4] = 8.873285624999166260485489e-05;
          A[1][5] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.003549314249999666395773978;
          A[2][2] = -0.001774657124999833197886989;
          A[2][3] = -0.0003992978531249624478405291;
          A[2][4] = -5.176083281249513313136690e-05;
          A[2][5] = -3.697202343749652185019147e-06;
          A[3][0] = -0.001597191412499849791362116;
          A[3][1] = -0.0007985957062499248956810582;
          A[3][2] = -0.0001715501887499838729045365;
          A[3][3] = -1.922545218749819288675887e-05;
          A[3][4] = -9.859206249999073763933813e-07;
          A[3][5] = -0.000000000000000000000000;
          A[4][1] = 4.225374107142459519017175e-06;
          A[4][2] = 2.112687053571229759508588e-06;
          A[4][3] = 5.089655174512508480514435e-07;
          A[4][4] = 7.842550426135625534185644e-08;
          A[4][5] = 7.202342228083738478251039e-09;
          A[5][2] = 1.067013663419813083379988e-08;
          A[5][3] = 5.335068317099065416899938e-09;
          A[5][4] = 8.891780528498442361499897e-10;
          A[5][5] = -0.000000000000000000000000;
          A[6][3] = -2.798112753723286118822838e-11;
          A[6][4] = -1.399056376861643059411419e-11;
          A[6][5] = -2.331760628102738432352365e-12;
        } // |m|=1
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][2] = -0.01745743121887938925640427;
          A[0][3] = -0.008728715609439694628202133;
          A[0][4] = -0.001879098499254378776962549;
          A[0][5] = -0.0002121562821738814756927088;
          A[0][6] = 4.546206046583174285521944e-05;
          A[1][1] = -0.05455447255899809749779550;
          A[1][2] = -0.02727723627949904874889775;
          A[1][3] = -0.004546206046583174502362379;
          A[1][4] = 0.000000000000000000000000;
          A[1][5] = 7.577010077638624712705051e-05;
          A[1][6] = 0.000000000000000000000000;
          A[2][0] = -0.03273268353539885572311974;
          A[2][1] = -0.01636634176769942786155987;
          A[2][2] = -0.002303411063602141782558075;
          A[2][3] = 0.0002121562821738814756927088;
          A[2][4] = 8.587278087990440618264276e-05;
          A[2][5] = 2.525670025879541457963957e-06;
          A[2][6] = -1.262835012939770728981979e-06;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.0001010268010351816583185583;
          A[3][2] = 5.051340051759082915927915e-05;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = -4.209450043132569378950911e-06;
          A[3][5] = -8.418900086265138546143586e-07;
          A[3][6] = 0.000000000000000000000000;
          A[4][0] = -0.0001731888017745971362904012;
          A[4][1] = -8.659440088729856814520058e-05;
          A[4][2] = -1.957558274603779727564105e-05;
          A[4][3] = -2.571591299077351433559303e-06;
          A[4][4] = -1.672846380777358556499320e-07;
          A[4][5] = 1.038695465188556009705100e-08;
          A[4][6] = 2.460068207025527434942638e-09;
          A[5][1] = 7.289090983779340214655036e-07;
          A[5][2] = 3.644545491889670107327518e-07;
          A[5][3] = 8.200227356751757344440222e-08;
          A[5][4] = 1.062992435134487136695342e-08;
          A[5][5] = 7.592803108103479252402227e-10;
          A[5][6] = 0.000000000000000000000000;
          A[6][2] = -7.645899633334972034770541e-10;
          A[6][3] = -3.822949816667486017385270e-10;
          A[6][4] = -8.601637087501843409869888e-11;
          A[6][5] = -1.115027029861350077633456e-11;
          A[6][6] = -7.964478784723929270202111e-13;
        } // |m|=0
        else if(abs(m)==1){
          A[0][2] = 0.01425393290199596790790437;
          A[0][3] = 0.007126966450997983953952186;
          A[0][4] = 0.001633263145020371395060854;
          A[0][5] = 0.0002227177015936869985610058;
          A[1][1] = 0.04454354031873740404900985;
          A[1][2] = 0.02227177015936870202450493;
          A[1][3] = 0.004454354031873740404900985;
          A[1][4] = 0.0003711961693228116461983068;
          A[1][5] = 0.000000000000000000000000;
          A[2][0] = 0.02672612419124243895995896;
          A[2][1] = 0.01336306209562121947997948;
          A[2][2] = 0.002227177015936870202450493;
          A[2][3] = 0.000000000000000000000000;
          A[2][4] = -4.949282257637488706327605e-05;
          A[2][5] = -6.186602822046860882909507e-06;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = -0.0004949282257637488977378148;
          A[3][2] = -0.0002474641128818744488689074;
          A[3][3] = -4.949282257637488706327605e-05;
          A[3][4] = -4.124401881364574204283987e-06;
          A[3][5] = 0.000000000000000000000000;
          A[4][0] = -0.0001060560483779461962445395;
          A[4][1] = -5.302802418897309812226976e-05;
          A[4][2] = -9.512906157536991152089655e-06;
          A[4][3] = -3.374510630207378932915669e-07;
          A[4][4] = 1.151618707134264178088394e-07;
          A[4][5] = 1.205182367931206667220669e-08;
          A[5][1] = 4.463638399745210251132709e-07;
          A[5][2] = 2.231819199872605125566354e-07;
          A[5][3] = 4.463638399745209854086014e-08;
          A[5][4] = 3.719698666454341958661985e-09;
          A[5][5] = 0.000000000000000000000000;
          A[6][2] = -4.682138181550919687664544e-10;
          A[6][3] = -2.341069090775459843832272e-10;
          A[6][4] = -4.682138181550919946158485e-11;
          A[6][5] = -3.901781817959099416603026e-12;
        } // |m|=1
      } // lb=3
    } // la=1
    else if(la==2){
      if(lb==0){
        if(abs(m)==0){
          A[0][2] = 0.000000000000000000000000;
          A[0][3] = 0.000000000000000000000000;
          A[0][4] = 0.0002661985687499749471569832;
          A[0][5] = 0.0001330992843749874735784916;
          A[0][6] = 2.218321406249791226308193e-05;
          A[1][1] = -0.03726779962499649628826504;
          A[1][2] = -0.01863389981249824814413252;
          A[1][3] = -0.003992978531249624478405291;
          A[1][4] = -0.0004436642812499582452616387;
          A[1][5] = -2.218321406249791226308193e-05;
          A[1][6] = 0.000000000000000000000000;
          A[2][0] = 0.03726779962499649628826504;
          A[2][1] = 0.01863389981249824814413252;
          A[2][2] = 0.003726779962499649368617982;
          A[2][3] = 0.0003105649968749707987882014;
          A[2][4] = -7.394404687499304370038294e-06;
          A[2][5] = -3.697202343749652185019147e-06;
          A[2][6] = -6.162003906249420308365245e-07;
          A[3][1] = 0.0002661985687499749471569832;
          A[3][2] = 0.0001330992843749874735784916;
          A[3][3] = 2.957761874999721748015317e-05;
          A[3][4] = 3.697202343749652185019147e-06;
          A[3][5] = 2.464801562499767911587861e-07;
          A[3][6] = 0.000000000000000000000000;
          A[4][2] = 0.000000000000000000000000;
          A[4][3] = 0.000000000000000000000000;
          A[4][4] = 1.440468445616747530214085e-08;
          A[4][5] = 7.202342228083737651070426e-09;
          A[4][6] = 1.200390371347289677443455e-09;
          A[5][3] = -2.667534158549532708449969e-09;
          A[5][4] = -1.333767079274766354224985e-09;
          A[5][5] = -2.222945132124610590374974e-10;
          A[5][6] = 0.000000000000000000000000;
          A[6][4] = -4.663521256205476864704731e-12;
          A[6][5] = -2.331760628102738432352365e-12;
          A[6][6] = -3.886267713504564053920609e-13;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][1] = -0.1721325931647740836716309;
          A[0][2] = -0.08606629658238704183581547;
          A[0][3] = -0.01659850005517464477389211;
          A[0][4] = -0.001127058645721735205894043;
          A[0][5] = 0.0001280748461047426321052223;
          A[0][6] = 3.842245383142278556580854e-05;
          A[1][0] = 0.2581988897471611532630220;
          A[1][1] = 0.1290994448735805766315110;
          A[1][2] = 0.02182395377624814225558936;
          A[1][3] = 0.0001536898153256911422632341;
          A[1][4] = -0.0003842245383142278963156668;
          A[1][5] = -2.561496922094852371053902e-05;
          A[1][6] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.001024598768837941056841778;
          A[2][2] = 0.0005122993844189705284208891;
          A[2][3] = 7.684490766284557113161707e-05;
          A[2][4] = -4.269161536824754234100820e-06;
          A[2][5] = -4.269161536824754234100820e-06;
          A[2][6] = -1.067290384206188558525205e-06;
          A[3][0] = 0.002766416675862440940208975;
          A[3][1] = 0.001383208337931220470104487;
          A[3][2] = 0.0003125026244955719919790815;
          A[3][3] = 4.098395075351763929211515e-05;
          A[3][4] = 3.557634614020628669589508e-06;
          A[3][5] = 2.846107691216502787440840e-07;
          A[3][6] = -0.000000000000000000000000;
          A[4][1] = 7.318562634556721181158714e-06;
          A[4][2] = 3.659281317278360590579357e-06;
          A[4][3] = 8.205661135715111588858272e-07;
          A[4][4] = 1.053429470125588635613005e-07;
          A[4][5] = 1.155076173383320953649549e-08;
          A[4][6] = 2.079137112089977766200024e-09;
          A[5][2] = -1.848121877413313459664829e-08;
          A[5][3] = -9.240609387066567298324144e-09;
          A[5][4] = -2.053468752681459353673109e-09;
          A[5][5] = -2.566835940851824192091386e-10;
          A[5][6] = -0.000000000000000000000000;
          A[6][3] = -4.846473454755192770035815e-11;
          A[6][4] = -2.423236727377596385017907e-11;
          A[6][5] = -5.384970505283547612016857e-12;
          A[6][6] = -6.731213131604434515021072e-13;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = 0.1490711984999859851530601;
          A[0][2] = 0.07453559924999299257653007;
          A[0][3] = 0.01650431126249844726583405;
          A[0][4] = 0.002040855693749807873993429;
          A[0][5] = 0.0001330992843749874735784916;
          A[1][0] = -0.2236067977499789638518024;
          A[1][1] = -0.1118033988749894819259012;
          A[1][2] = -0.02342547404999779578349539;
          A[1][3] = -0.002395787118749774687043175;
          A[1][4] = -8.873285624999166260485489e-05;
          A[1][5] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.003549314249999666395773978;
          A[2][2] = -0.001774657124999833197886989;
          A[2][3] = -0.0003992978531249624478405291;
          A[2][4] = -5.176083281249513313136690e-05;
          A[2][5] = -3.697202343749652185019147e-06;
          A[3][0] = 0.001597191412499849791362116;
          A[3][1] = 0.0007985957062499248956810582;
          A[3][2] = 0.0001715501887499838729045365;
          A[3][3] = 1.922545218749819288675887e-05;
          A[3][4] = 9.859206249999073763933813e-07;
          A[3][5] = -0.000000000000000000000000;
          A[4][1] = 4.225374107142459519017175e-06;
          A[4][2] = 2.112687053571229759508588e-06;
          A[4][3] = 5.089655174512508480514435e-07;
          A[4][4] = 7.842550426135625534185644e-08;
          A[4][5] = 7.202342228083738478251039e-09;
          A[5][2] = -1.067013663419813083379988e-08;
          A[5][3] = -5.335068317099065416899938e-09;
          A[5][4] = -8.891780528498442361499897e-10;
          A[5][5] = -0.000000000000000000000000;
          A[6][3] = -2.798112753723286118822838e-11;
          A[6][4] = -1.399056376861643059411419e-11;
          A[6][5] = -2.331760628102738432352365e-12;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.05952380952380952050528862;
          A[0][3] = -0.01190476190476190410105772;
          A[0][4] = -0.003571428571428571317053491;
          A[0][5] = -9.920634920634920626315856e-05;
          A[0][6] = 4.960317460317460313157928e-05;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[1][5] = 0.000000000000000000000000;
          A[1][6] = 0.000000000000000000000000;
          A[2][0] = 0.02380952380952380820211545;
          A[2][1] = 0.01190476190476190410105772;
          A[2][2] = 0.002579365079365079308632014;
          A[2][3] = 0.0002976190476190476458945300;
          A[2][4] = 1.653439153439153550657036e-05;
          A[2][5] = 0.000000000000000000000000;
          A[2][6] = -1.377865961199294625547530e-06;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[3][5] = 0.000000000000000000000000;
          A[3][6] = 0.000000000000000000000000;
          A[4][0] = 0.0001700680272108843419775343;
          A[4][1] = 8.503401360544217098876713e-05;
          A[4][2] = 1.932591218305504163319126e-05;
          A[4][3] = 2.576788291074005296982284e-06;
          A[4][4] = 2.183112302159921201445155e-07;
          A[4][5] = 1.252605419272086011193309e-08;
          A[4][6] = 2.684154469868755738271376e-09;
          A[5][1] = 0.000000000000000000000000;
          A[5][2] = 0.000000000000000000000000;
          A[5][3] = 0.000000000000000000000000;
          A[5][4] = 0.000000000000000000000000;
          A[5][5] = 0.000000000000000000000000;
          A[5][6] = 0.000000000000000000000000;
          A[6][2] = -7.508124391241274293456103e-10;
          A[6][3] = -3.754062195620637146728051e-10;
          A[6][4] = -8.342360434712526561905767e-11;
          A[6][5] = -1.042795054339065820238221e-11;
          A[6][6] = -8.689958786158882508479814e-13;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = -1.000000000000000000000000;
          A[0][1] = -0.5000000000000000000000000;
          A[0][2] = -0.07142857142857142460634634;
          A[0][3] = 0.005952380952380952050528862;
          A[0][4] = 0.002976190476190476025264431;
          A[0][5] = 0.0002976190476190476458945300;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[1][5] = 0.000000000000000000000000;
          A[2][0] = -0.01190476190476190410105772;
          A[2][1] = -0.005952380952380952050528862;
          A[2][2] = -0.001984126984126984016842954;
          A[2][3] = -0.0004960317460317460042107385;
          A[2][4] = -8.267195767195766736845641e-05;
          A[2][5] = -8.267195767195767753285178e-06;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[3][5] = 0.000000000000000000000000;
          A[4][0] = 0.0001133786848072562415375500;
          A[4][1] = 5.668934240362812076877500e-05;
          A[4][2] = 1.288394145537002606139495e-05;
          A[4][3] = 1.717858860716003672493681e-06;
          A[4][4] = 1.610492681921253310613928e-07;
          A[4][5] = 1.610492681921253442962826e-08;
          A[5][1] = 0.000000000000000000000000;
          A[5][2] = 0.000000000000000000000000;
          A[5][3] = 0.000000000000000000000000;
          A[5][4] = 0.000000000000000000000000;
          A[5][5] = 0.000000000000000000000000;
          A[6][2] = -5.005416260827516195637402e-10;
          A[6][3] = -2.502708130413758097818701e-10;
          A[6][4] = -5.213975271695329585867245e-11;
          A[6][5] = -5.213975271695329101191105e-12;
        } // |m|=1
        else if(abs(m)==2){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.1071428571428571369095195;
          A[0][3] = 0.01190476190476190410105772;
          A[0][4] = 0.0005952380952380952917890600;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[2][0] = -0.02380952380952380820211545;
          A[2][1] = -0.01190476190476190410105772;
          A[2][2] = -0.002579365079365079308632014;
          A[2][3] = -0.0002976190476190476458945300;
          A[2][4] = -1.653439153439153550657036e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[4][0] = 2.834467120181406038438750e-05;
          A[4][1] = 1.417233560090703019219375e-05;
          A[4][2] = 3.220985363842506515348737e-06;
          A[4][3] = 4.294647151790009181234202e-07;
          A[4][4] = 3.220985363842506885925652e-08;
          A[5][1] = 0.000000000000000000000000;
          A[5][2] = 0.000000000000000000000000;
          A[5][3] = 0.000000000000000000000000;
          A[5][4] = 0.000000000000000000000000;
          A[6][2] = -1.251354065206879048909350e-10;
          A[6][3] = -6.256770326034395244546752e-11;
          A[6][4] = -1.042795054339065820238221e-11;
        } // |m|=2
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][1] = 0.2535462764185549833229061;
          A[0][2] = 0.1267731382092774916614530;
          A[0][3] = 0.01690308509457033245282354;
          A[0][4] = -0.002112885636821291556602942;
          A[0][5] = -0.0005869126768948032101674839;
          A[0][6] = 5.869126768948031424048481e-05;
          A[1][0] = 0.2535462764185549833229061;
          A[1][1] = 0.1267731382092774916614530;
          A[1][2] = 0.01690308509457033245282354;
          A[1][3] = -0.002112885636821291556602942;
          A[1][4] = -0.0005869126768948032101674839;
          A[1][5] = 5.869126768948031424048481e-05;
          A[1][6] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.002347650707579212840669936;
          A[2][2] = -0.001173825353789606420334968;
          A[2][3] = -7.825502358597375232064641e-05;
          A[2][4] = 5.869126768948031424048481e-05;
          A[2][5] = 9.781877948246719040080802e-06;
          A[2][6] = -1.630312991374453243932879e-06;
          A[3][0] = 0.001878120566063370055695514;
          A[3][1] = 0.0009390602830316850278477570;
          A[3][2] = 0.0002034630613235317560336807;
          A[3][3] = 2.347650707579212569619392e-05;
          A[3][4] = -0.000000000000000000000000;
          A[3][5] = -6.521251965497813399247990e-07;
          A[3][6] = -0.000000000000000000000000;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 1.016299007610048811190069e-06;
          A[4][2] = 5.081495038050244055950343e-07;
          A[4][3] = 1.524448511415073110905984e-07;
          A[4][4] = 3.387663358700162615734297e-08;
          A[4][5] = 1.058644799593800734698906e-08;
          A[4][6] = 3.175934398781402452250903e-09;
          A[5][0] = 8.469158396750407254019791e-06;
          A[5][1] = 4.234579198375203627009895e-06;
          A[5][2] = 9.598379516317127543596072e-07;
          A[5][3] = 1.270373759512561013987586e-07;
          A[5][4] = 1.058644799593800734698906e-08;
          A[5][5] = 5.881359997743338219197298e-10;
          A[5][6] = -0.000000000000000000000000;
          A[6][1] = -1.480622097333987092440354e-08;
          A[6][2] = -7.403110486669935462201772e-09;
          A[6][3] = -1.678038376978518859172783e-09;
          A[6][4] = -2.220933146000980721378593e-10;
          A[6][5] = -1.850777621667483826776352e-11;
          A[6][6] = -1.028209789815268837530949e-12;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = -0.2390457218668787242599194;
          A[0][2] = -0.1195228609334393621299597;
          A[0][3] = -0.01792842914001590431949396;
          A[0][4] = 0.0009960238411119946844163309;
          A[0][5] = 0.0004980119205559973422081654;
          A[1][0] = -0.2390457218668787242599194;
          A[1][1] = -0.1195228609334393621299597;
          A[1][2] = -0.01792842914001590431949396;
          A[1][3] = 0.0009960238411119946844163309;
          A[1][4] = 0.0004980119205559973422081654;
          A[1][5] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.003320079470373315903841682;
          A[2][2] = 0.001660039735186657951920841;
          A[2][3] = 0.0001660039735186657897710732;
          A[2][4] = -5.533465783955525873951536e-05;
          A[2][5] = -1.383366445988881468487884e-05;
          A[3][0] = -0.0006640158940746631590842930;
          A[3][1] = -0.0003320079470373315795421465;
          A[3][2] = -0.0001328031788149326372378695;
          A[3][3] = -3.873426048768868247291347e-05;
          A[3][4] = -5.533465783955526551577894e-06;
          A[3][5] = -0.000000000000000000000000;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 8.623583039930690047705254e-06;
          A[4][2] = 4.311791519965345023852627e-06;
          A[4][3] = 1.185742667990469945086944e-06;
          A[4][4] = 2.335553739981228519293800e-07;
          A[4][5] = 2.694869699978340573733443e-08;
          A[5][0] = 5.988599333285201657304467e-06;
          A[5][1] = 2.994299666642600828652234e-06;
          A[5][2] = 6.587459266613722034793151e-07;
          A[5][3] = 7.984799111046935719537821e-08;
          A[5][4] = 4.990499444404334824711138e-09;
          A[5][5] = -0.000000000000000000000000;
          A[6][1] = -1.046957925399510815112808e-08;
          A[6][2] = -5.234789626997554075564038e-09;
          A[6][3] = -1.151653717939461739459772e-09;
          A[6][4] = -1.395943900532680990312672e-10;
          A[6][5] = -8.724649378329256189454200e-12;
        } // |m|=1
        else if(abs(m)==2){
          A[0][1] = 0.1889822365046136265487320;
          A[0][2] = 0.09449111825230681327436599;
          A[0][3] = 0.01889822365046136057320503;
          A[0][4] = 0.001574851970871780264607520;
          A[1][0] = 0.1889822365046136265487320;
          A[1][1] = 0.09449111825230681327436599;
          A[1][2] = 0.01889822365046136057320503;
          A[1][3] = 0.001574851970871780264607520;
          A[1][4] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.005249506569572601026585357;
          A[2][2] = -0.002624753284786300513292678;
          A[2][3] = -0.0005249506569572600159223619;
          A[2][4] = -4.374588807977167025228468e-05;
          A[3][0] = -0.002099802627829040063689447;
          A[3][1] = -0.001049901313914520031844724;
          A[3][2] = -0.0002099802627829040172109665;
          A[3][3] = -1.749835523190866810091387e-05;
          A[3][4] = -0.000000000000000000000000;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 1.022631149916740433763410e-05;
          A[4][2] = 5.113155749583702168817052e-06;
          A[4][3] = 1.022631149916740349060116e-06;
          A[4][4] = 8.521926249306170457826950e-08;
          A[5][0] = 1.893761388734704405011609e-06;
          A[5][1] = 9.468806943673522025058044e-07;
          A[5][2] = 1.893761388734704457951168e-07;
          A[5][3] = 1.578134490612253604668558e-08;
          A[5][4] = -0.000000000000000000000000;
          A[6][1] = -3.310771658627105678365386e-09;
          A[6][2] = -1.655385829313552839182693e-09;
          A[6][3] = -3.310771658627105471570233e-10;
          A[6][4] = -2.758976382189254775053478e-11;
        } // |m|=2
      } // lb=3
    } // la=2
    else if(la==3){
      if(lb==0){
        if(abs(m)==0){
          A[0][3] = -0.001259881576697424211686016;
          A[0][4] = -0.0006299407883487121058430080;
          A[0][5] = -5.249506569572601107900520e-05;
          A[0][6] = 2.624753284786300553950260e-05;
          A[1][2] = -0.001889822365046136317529024;
          A[1][3] = -0.0009449111825230681587645121;
          A[1][4] = -0.0002624753284786300621712896;
          A[1][5] = -5.249506569572601107900520e-05;
          A[1][6] = -0.000000000000000000000000;
          A[2][1] = 0.006299407883487121058430080;
          A[2][2] = 0.003149703941743560529215040;
          A[2][3] = 0.0006474391435806207874964491;
          A[2][4] = 6.124424331168034512946213e-05;
          A[2][5] = -0.000000000000000000000000;
          A[2][6] = -7.290981346628612414574903e-07;
          A[3][0] = -0.003149703941743560529215040;
          A[3][1] = -0.001574851970871780264607520;
          A[3][2] = -0.0002764740126641569641259555;
          A[3][3] = -6.999342092763467917991907e-06;
          A[3][4] = 4.374588807977167025228468e-06;
          A[3][5] = 5.832785077302889931659922e-07;
          A[3][6] = -0.000000000000000000000000;
          A[4][1] = -2.499765033129809873907630e-05;
          A[4][2] = -1.249882516564904936953815e-05;
          A[4][3] = -2.454314759800177007150867e-06;
          A[4][4] = -1.855886160960010384674012e-07;
          A[4][5] = 6.628164860571465305044068e-09;
          A[4][6] = 1.420321041551028368278794e-09;
          A[5][2] = -6.312537962449014418674233e-08;
          A[5][3] = -3.156268981224507209337117e-08;
          A[5][4] = -6.312537962449015245854846e-09;
          A[5][5] = -5.260448302040845693553783e-10;
          A[5][6] = -0.000000000000000000000000;
          A[6][3] = -5.517952764378509550106957e-11;
          A[6][4] = -2.758976382189254775053478e-11;
          A[6][5] = -5.517952764378509711665670e-12;
          A[6][6] = -4.598293970315424423140739e-13;
        } // |m|=0
      } // lb=0
      else if(lb==1){
        if(abs(m)==0){
          A[0][2] = -0.01745743121887938925640427;
          A[0][3] = -0.008728715609439694628202133;
          A[0][4] = -0.001879098499254378776962549;
          A[0][5] = -0.0002121562821738814756927088;
          A[0][6] = 4.546206046583174285521944e-05;
          A[1][1] = 0.05455447255899809749779550;
          A[1][2] = 0.02727723627949904874889775;
          A[1][3] = 0.004546206046583174502362379;
          A[1][4] = 0.000000000000000000000000;
          A[1][5] = -7.577010077638624712705051e-05;
          A[1][6] = 0.000000000000000000000000;
          A[2][0] = -0.03273268353539885572311974;
          A[2][1] = -0.01636634176769942786155987;
          A[2][2] = -0.002303411063602141782558075;
          A[2][3] = 0.0002121562821738814756927088;
          A[2][4] = 8.587278087990440618264276e-05;
          A[2][5] = 2.525670025879541457963957e-06;
          A[2][6] = -1.262835012939770728981979e-06;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = -0.0001010268010351816583185583;
          A[3][2] = -5.051340051759082915927915e-05;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 4.209450043132569378950911e-06;
          A[3][5] = 8.418900086265138546143586e-07;
          A[3][6] = 0.000000000000000000000000;
          A[4][0] = -0.0001731888017745971362904012;
          A[4][1] = -8.659440088729856814520058e-05;
          A[4][2] = -1.957558274603779727564105e-05;
          A[4][3] = -2.571591299077351433559303e-06;
          A[4][4] = -1.672846380777358556499320e-07;
          A[4][5] = 1.038695465188556009705100e-08;
          A[4][6] = 2.460068207025527434942638e-09;
          A[5][1] = -7.289090983779340214655036e-07;
          A[5][2] = -3.644545491889670107327518e-07;
          A[5][3] = -8.200227356751757344440222e-08;
          A[5][4] = -1.062992435134487136695342e-08;
          A[5][5] = -7.592803108103479252402227e-10;
          A[5][6] = 0.000000000000000000000000;
          A[6][2] = -7.645899633334972034770541e-10;
          A[6][3] = -3.822949816667486017385270e-10;
          A[6][4] = -8.601637087501843409869888e-11;
          A[6][5] = -1.115027029861350077633456e-11;
          A[6][6] = -7.964478784723929270202111e-13;
        } // |m|=0
        else if(abs(m)==1){
          A[0][2] = 0.01425393290199596790790437;
          A[0][3] = 0.007126966450997983953952186;
          A[0][4] = 0.001633263145020371395060854;
          A[0][5] = 0.0002227177015936869985610058;
          A[1][1] = -0.04454354031873740404900985;
          A[1][2] = -0.02227177015936870202450493;
          A[1][3] = -0.004454354031873740404900985;
          A[1][4] = -0.0003711961693228116461983068;
          A[1][5] = 0.000000000000000000000000;
          A[2][0] = 0.02672612419124243895995896;
          A[2][1] = 0.01336306209562121947997948;
          A[2][2] = 0.002227177015936870202450493;
          A[2][3] = 0.000000000000000000000000;
          A[2][4] = -4.949282257637488706327605e-05;
          A[2][5] = -6.186602822046860882909507e-06;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.0004949282257637488977378148;
          A[3][2] = 0.0002474641128818744488689074;
          A[3][3] = 4.949282257637488706327605e-05;
          A[3][4] = 4.124401881364574204283987e-06;
          A[3][5] = 0.000000000000000000000000;
          A[4][0] = -0.0001060560483779461962445395;
          A[4][1] = -5.302802418897309812226976e-05;
          A[4][2] = -9.512906157536991152089655e-06;
          A[4][3] = -3.374510630207378932915669e-07;
          A[4][4] = 1.151618707134264178088394e-07;
          A[4][5] = 1.205182367931206667220669e-08;
          A[5][1] = -4.463638399745210251132709e-07;
          A[5][2] = -2.231819199872605125566354e-07;
          A[5][3] = -4.463638399745209854086014e-08;
          A[5][4] = -3.719698666454341958661985e-09;
          A[5][5] = 0.000000000000000000000000;
          A[6][2] = -4.682138181550919687664544e-10;
          A[6][3] = -2.341069090775459843832272e-10;
          A[6][4] = -4.682138181550919946158485e-11;
          A[6][5] = -3.901781817959099416603026e-12;
        } // |m|=1
      } // lb=1
      else if(lb==2){
        if(abs(m)==0){
          A[0][1] = 0.2535462764185549833229061;
          A[0][2] = 0.1267731382092774916614530;
          A[0][3] = 0.01690308509457033245282354;
          A[0][4] = -0.002112885636821291556602942;
          A[0][5] = -0.0005869126768948032101674839;
          A[0][6] = 5.869126768948031424048481e-05;
          A[1][0] = -0.2535462764185549833229061;
          A[1][1] = -0.1267731382092774916614530;
          A[1][2] = -0.01690308509457033245282354;
          A[1][3] = 0.002112885636821291556602942;
          A[1][4] = 0.0005869126768948032101674839;
          A[1][5] = -5.869126768948031424048481e-05;
          A[1][6] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.002347650707579212840669936;
          A[2][2] = -0.001173825353789606420334968;
          A[2][3] = -7.825502358597375232064641e-05;
          A[2][4] = 5.869126768948031424048481e-05;
          A[2][5] = 9.781877948246719040080802e-06;
          A[2][6] = -1.630312991374453243932879e-06;
          A[3][0] = -0.001878120566063370055695514;
          A[3][1] = -0.0009390602830316850278477570;
          A[3][2] = -0.0002034630613235317560336807;
          A[3][3] = -2.347650707579212569619392e-05;
          A[3][4] = -0.000000000000000000000000;
          A[3][5] = 6.521251965497813399247990e-07;
          A[3][6] = -0.000000000000000000000000;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 1.016299007610048811190069e-06;
          A[4][2] = 5.081495038050244055950343e-07;
          A[4][3] = 1.524448511415073110905984e-07;
          A[4][4] = 3.387663358700162615734297e-08;
          A[4][5] = 1.058644799593800734698906e-08;
          A[4][6] = 3.175934398781402452250903e-09;
          A[5][0] = -8.469158396750407254019791e-06;
          A[5][1] = -4.234579198375203627009895e-06;
          A[5][2] = -9.598379516317127543596072e-07;
          A[5][3] = -1.270373759512561013987586e-07;
          A[5][4] = -1.058644799593800734698906e-08;
          A[5][5] = -5.881359997743338219197298e-10;
          A[5][6] = -0.000000000000000000000000;
          A[6][1] = -1.480622097333987092440354e-08;
          A[6][2] = -7.403110486669935462201772e-09;
          A[6][3] = -1.678038376978518859172783e-09;
          A[6][4] = -2.220933146000980721378593e-10;
          A[6][5] = -1.850777621667483826776352e-11;
          A[6][6] = -1.028209789815268837530949e-12;
        } // |m|=0
        else if(abs(m)==1){
          A[0][1] = -0.2390457218668787242599194;
          A[0][2] = -0.1195228609334393621299597;
          A[0][3] = -0.01792842914001590431949396;
          A[0][4] = 0.0009960238411119946844163309;
          A[0][5] = 0.0004980119205559973422081654;
          A[1][0] = 0.2390457218668787242599194;
          A[1][1] = 0.1195228609334393621299597;
          A[1][2] = 0.01792842914001590431949396;
          A[1][3] = -0.0009960238411119946844163309;
          A[1][4] = -0.0004980119205559973422081654;
          A[1][5] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = 0.003320079470373315903841682;
          A[2][2] = 0.001660039735186657951920841;
          A[2][3] = 0.0001660039735186657897710732;
          A[2][4] = -5.533465783955525873951536e-05;
          A[2][5] = -1.383366445988881468487884e-05;
          A[3][0] = 0.0006640158940746631590842930;
          A[3][1] = 0.0003320079470373315795421465;
          A[3][2] = 0.0001328031788149326372378695;
          A[3][3] = 3.873426048768868247291347e-05;
          A[3][4] = 5.533465783955526551577894e-06;
          A[3][5] = -0.000000000000000000000000;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 8.623583039930690047705254e-06;
          A[4][2] = 4.311791519965345023852627e-06;
          A[4][3] = 1.185742667990469945086944e-06;
          A[4][4] = 2.335553739981228519293800e-07;
          A[4][5] = 2.694869699978340573733443e-08;
          A[5][0] = -5.988599333285201657304467e-06;
          A[5][1] = -2.994299666642600828652234e-06;
          A[5][2] = -6.587459266613722034793151e-07;
          A[5][3] = -7.984799111046935719537821e-08;
          A[5][4] = -4.990499444404334824711138e-09;
          A[5][5] = -0.000000000000000000000000;
          A[6][1] = -1.046957925399510815112808e-08;
          A[6][2] = -5.234789626997554075564038e-09;
          A[6][3] = -1.151653717939461739459772e-09;
          A[6][4] = -1.395943900532680990312672e-10;
          A[6][5] = -8.724649378329256189454200e-12;
        } // |m|=1
        else if(abs(m)==2){
          A[0][1] = 0.1889822365046136265487320;
          A[0][2] = 0.09449111825230681327436599;
          A[0][3] = 0.01889822365046136057320503;
          A[0][4] = 0.001574851970871780264607520;
          A[1][0] = -0.1889822365046136265487320;
          A[1][1] = -0.09449111825230681327436599;
          A[1][2] = -0.01889822365046136057320503;
          A[1][3] = -0.001574851970871780264607520;
          A[1][4] = -0.000000000000000000000000;
          A[2][0] = -0.000000000000000000000000;
          A[2][1] = -0.005249506569572601026585357;
          A[2][2] = -0.002624753284786300513292678;
          A[2][3] = -0.0005249506569572600159223619;
          A[2][4] = -4.374588807977167025228468e-05;
          A[3][0] = 0.002099802627829040063689447;
          A[3][1] = 0.001049901313914520031844724;
          A[3][2] = 0.0002099802627829040172109665;
          A[3][3] = 1.749835523190866810091387e-05;
          A[3][4] = -0.000000000000000000000000;
          A[4][0] = -0.000000000000000000000000;
          A[4][1] = 1.022631149916740433763410e-05;
          A[4][2] = 5.113155749583702168817052e-06;
          A[4][3] = 1.022631149916740349060116e-06;
          A[4][4] = 8.521926249306170457826950e-08;
          A[5][0] = -1.893761388734704405011609e-06;
          A[5][1] = -9.468806943673522025058044e-07;
          A[5][2] = -1.893761388734704457951168e-07;
          A[5][3] = -1.578134490612253604668558e-08;
          A[5][4] = -0.000000000000000000000000;
          A[6][1] = -3.310771658627105678365386e-09;
          A[6][2] = -1.655385829313552839182693e-09;
          A[6][3] = -3.310771658627105471570233e-10;
          A[6][4] = -2.758976382189254775053478e-11;
        } // |m|=2
      } // lb=2
      else if(lb==3){
        if(abs(m)==0){
          A[0][0] = -0.9999999999999998889776975;
          A[0][1] = -0.4999999999999999444888488;
          A[0][2] = -0.01000000000000000020816682;
          A[0][3] = 0.03666666666666666685170384;
          A[0][4] = 0.005000000000000000104083409;
          A[0][5] = -0.001250000000000000026020852;
          A[0][6] = 6.944444444444444438421099e-05;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[1][5] = 0.000000000000000000000000;
          A[1][6] = 0.000000000000000000000000;
          A[2][0] = -0.02222222222222221960086230;
          A[2][1] = -0.01111111111111110980043115;
          A[2][2] = -0.003055555555555555281854740;
          A[2][3] = -0.0006018518518518517880847019;
          A[2][4] = -2.314814814814814699869307e-05;
          A[2][5] = 2.314814814814814699869307e-05;
          A[2][6] = -1.929012345679012391063247e-06;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[3][5] = 0.000000000000000000000000;
          A[3][6] = 0.000000000000000000000000;
          A[4][0] = -0.0001082251082251082250143548;
          A[4][1] = -5.411255411255411250717740e-05;
          A[4][2] = -1.046176046176046107376127e-05;
          A[4][3] = -7.215007215007214789198749e-07;
          A[4][4] = 4.509379509379509243249218e-08;
          A[4][5] = 7.515632515632515405415364e-09;
          A[4][6] = 3.757816257816257702707682e-09;
          A[5][0] = 0.000000000000000000000000;
          A[5][1] = 0.000000000000000000000000;
          A[5][2] = 0.000000000000000000000000;
          A[5][3] = 0.000000000000000000000000;
          A[5][4] = 0.000000000000000000000000;
          A[5][5] = 0.000000000000000000000000;
          A[5][6] = 0.000000000000000000000000;
          A[6][0] = -3.503791382579261262499926e-07;
          A[6][1] = -1.751895691289630631249963e-07;
          A[6][2] = -4.029360089966150134237560e-08;
          A[6][3] = -5.547669689083829847012257e-09;
          A[6][4] = -4.963704458653953543440827e-10;
          A[6][5] = -2.919826152149384361290504e-11;
          A[6][6] = -1.216594230062243551187174e-12;
        } // |m|=0
        else if(abs(m)==1){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.02000000000000000041633363;
          A[0][3] = -0.03166666666666666934970564;
          A[0][4] = -0.005000000000000000104083409;
          A[0][5] = 0.0008333333333333333868206405;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[1][5] = 0.000000000000000000000000;
          A[2][0] = 0.01666666666666666643537020;
          A[2][1] = 0.008333333333333333217685102;
          A[2][2] = 0.002777777777777777883788657;
          A[2][3] = 0.0006944444444444444709471642;
          A[2][4] = 4.629629629629629399738613e-05;
          A[2][5] = -2.314814814814814699869307e-05;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[3][5] = 0.000000000000000000000000;
          A[4][0] = 1.803751803751803750239247e-05;
          A[4][1] = 9.018759018759018751196233e-06;
          A[4][2] = 4.689754689754690089435220e-06;
          A[4][3] = 1.593314093314093411531845e-06;
          A[4][4] = 3.306878306878306910731658e-07;
          A[4][5] = 4.509379509379509243249218e-08;
          A[5][0] = 0.000000000000000000000000;
          A[5][1] = 0.000000000000000000000000;
          A[5][2] = 0.000000000000000000000000;
          A[5][3] = 0.000000000000000000000000;
          A[5][4] = 0.000000000000000000000000;
          A[5][5] = 0.000000000000000000000000;
          A[6][0] = -2.627843536934445814526047e-07;
          A[6][1] = -1.313921768467222907263023e-07;
          A[6][2] = -2.978222675192372126064496e-08;
          A[6][3] = -3.941765305401669185010213e-09;
          A[6][4] = -3.211808767364323185160467e-10;
          A[6][5] = -1.459913076074692180645252e-11;
        } // |m|=1
        else if(abs(m)==2){
          A[0][0] = -1.000000000000000000000000;
          A[0][1] = -0.5000000000000000000000000;
          A[0][2] = -0.05000000000000000277555756;
          A[0][3] = 0.01666666666666666643537020;
          A[0][4] = 0.004166666666666666608842551;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[1][4] = 0.000000000000000000000000;
          A[2][0] = 0.000000000000000000000000;
          A[2][1] = 0.000000000000000000000000;
          A[2][2] = -0.001388888888888888941894328;
          A[2][3] = -0.0006944444444444444709471642;
          A[2][4] = -0.0001157407407407407451578607;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[3][4] = 0.000000000000000000000000;
          A[4][0] = 0.0001262626262626262625167473;
          A[4][1] = 6.313131313131313125837363e-05;
          A[4][2] = 1.533189033189033187703360e-05;
          A[4][3] = 2.405002405002405141491153e-06;
          A[4][4] = 2.254689754689754952496854e-07;
          A[5][0] = 0.000000000000000000000000;
          A[5][1] = 0.000000000000000000000000;
          A[5][2] = 0.000000000000000000000000;
          A[5][3] = 0.000000000000000000000000;
          A[5][4] = 0.000000000000000000000000;
          A[6][0] = -1.051137414773778537568655e-07;
          A[6][1] = -5.255687073868892687843277e-08;
          A[6][2] = -1.138732199338259950017145e-08;
          A[6][3] = -1.313921768467223130601789e-09;
          A[6][4] = -7.299565380373461549461113e-11;
        } // |m|=2
        else if(abs(m)==3){
          A[0][0] = 1.000000000000000000000000;
          A[0][1] = 0.5000000000000000000000000;
          A[0][2] = 0.1000000000000000055511151;
          A[0][3] = 0.008333333333333333217685102;
          A[1][0] = 0.000000000000000000000000;
          A[1][1] = 0.000000000000000000000000;
          A[1][2] = 0.000000000000000000000000;
          A[1][3] = 0.000000000000000000000000;
          A[2][0] = -0.02777777777777777623580135;
          A[2][1] = -0.01388888888888888811790068;
          A[2][2] = -0.002777777777777777883788657;
          A[2][3] = -0.0002314814814814814903157214;
          A[3][0] = 0.000000000000000000000000;
          A[3][1] = 0.000000000000000000000000;
          A[3][2] = 0.000000000000000000000000;
          A[3][3] = 0.000000000000000000000000;
          A[4][0] = 5.411255411255411250717740e-05;
          A[4][1] = 2.705627705627705625358870e-05;
          A[4][2] = 5.411255411255411250717740e-06;
          A[4][3] = 4.509379509379509375598116e-07;
          A[5][0] = 0.000000000000000000000000;
          A[5][1] = 0.000000000000000000000000;
          A[5][2] = 0.000000000000000000000000;
          A[5][3] = 0.000000000000000000000000;
          A[6][0] = -1.751895691289630565075514e-08;
          A[6][1] = -8.759478456448152825377570e-09;
          A[6][2] = -1.751895691289630771870667e-09;
          A[6][3] = -1.459913076074692309892223e-10;
        } // |m|=3
      } // lb=3
    } // la=3
  } // nb=3
} // na=3


} // A_coefficients_fast



}// namespace libmolint
}// namespace liblibra
